#define MAX_OID_LEN	    128	/* max subid's in an oid */

#define ASN_BOOLEAN	    (0x01)
#define ASN_INTEGER	    (0x02)
#define ASN_BIT_STR	    (0x03)
#define ASN_OCTET_STR	    (0x04)
#define ASN_NULL	    (0x05)
#define ASN_OBJECT_ID	    (0x06)
#define ASN_SEQUENCE	    (0x10)
#define ASN_SET		    (0x11)

#define ASN_UNIVERSAL	    (0x00)
#define ASN_APPLICATION     (0x40)
#define ASN_CONTEXT	    (0x80)
#define ASN_PRIVATE	    (0xC0)

#define ASN_PRIMITIVE	    (0x00)
#define ASN_CONSTRUCTOR	    (0x20)

#define ASN_LONG_LEN	    (0x80)
#define ASN_EXTENSION_ID    (0x1F)
#define ASN_BIT8	    (0x80)
#define SID_MAX_LEN	200
#define MAX_NAME_LEN	128  /* number of subid's in a objid */

#ifndef NULL
#define NULL 0


#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

#define READ	    1
#define WRITE	    0

#define RESERVE1    0
#define RESERVE2    1
#define COMMIT      2
#define ACTION      3
#define FREE        4

#define RONLY	0xAAAA	/* read access for everyone */
#define RWRITE	0xAABA	/* add write access for community private */
#define NOACCESS 0x0000	/* no access for anybody */

#define INTEGER	    ASN_INTEGER
#define STRING	    ASN_OCTET_STR
#define OBJID	    ASN_OBJECT_ID
#define NULLOBJ	    ASN_NULL

/* defined types (from the SMI, RFC 1065) */
#define IPADDRESS   (ASN_APPLICATION | 0)
#define COUNTER	    (ASN_APPLICATION | 1)
#define GAUGE	    (ASN_APPLICATION | 2)
#define TIMETICKS   (ASN_APPLICATION | 3)
#define OPAQUE	    (ASN_APPLICATION | 4)

#define NSAP        (ASN_APPLICATION | 5)
#define COUNTER64   (ASN_APPLICATION | 6)
#define UINTEGER    (ASN_APPLICATION | 7)

#ifdef DEBUG
#define ERROR(string)	printf("%s(%d): %s\n",__FILE__, __LINE__, string);
#else
#define ERROR(string)
#endif
/*
 * asn_parse_int - pulls a long out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_int(data, datalength, type, intp, intsize)
	u_char  *data;	  /* IN - pointer to start of object */
	int	    *datalength;  /* IN/OUT - number of valid bytes left in buffer */
	u_char  *type;	  /* OUT - asn type of object */
	long    *intp;	  /* IN/OUT - pointer to start of output buffer */
	int     intsize;      /* IN - size of output buffer */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */
	u_char *bufp = data;
	u_long	    asn_length;
	long   value = 0;

	if (intsize != sizeof (long)){
		ERROR("not long");
		return NULL;
	}
	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL){
		ERROR("bad length");
		return NULL;
	}
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	if (asn_length > intsize){
		ERROR("I don't support such large integers");
		return NULL;
	}
	*datalength -= (int)asn_length + (bufp - data);
	if (*bufp & 0x80)
		value = -1; /* integer is negative */
	while(asn_length--)
		value = (value << 8) | *bufp++;
	*intp = value;
	return bufp;
}


/*
 * asn_parse_unsigned_int - pulls an unsigned long out of an ASN int type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_unsigned_int(data, datalength, type, intp, intsize)
	u_char	    *data;	/* IN - pointer to start of object */
	int	    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char		    *type;	/* OUT - asn type of object */
	u_long		    *intp;	/* IN/OUT - pointer to start of output buffer */
	int			    intsize;    /* IN - size of output buffer */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */
	u_char *bufp = data;
	u_long	    asn_length;
	u_long value = 0;

	if (intsize != sizeof (long)){
		ERROR("not long");
		return NULL;
	}
	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL){
		ERROR("bad length");
		return NULL;
	}
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	if ((asn_length > (intsize + 1)) ||
			((asn_length == intsize + 1) && *bufp != 0x00)){
		ERROR("I don't support such large integers");
		return NULL;
	}
	*datalength -= (int)asn_length + (bufp - data);
	if (*bufp & 0x80)
		value = -1; /* integer is negative */
	while(asn_length--)
		value = (value << 8) | *bufp++;
	*intp = value;
	return bufp;
}


/*
 * asn_build_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_int(data, datalength, type, intp, intsize)
	u_char *data;	/* IN - pointer to start of output buffer */
	int    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	/* IN - asn type of object */
	long   *intp;	/* IN - pointer to start of long integer */
	int    intsize;    /* IN - size of *intp */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */

	long integer;
	u_long mask;

	if (intsize != sizeof (long)) {
		ERROR("not long");
		return NULL;
	}
	integer = *intp;
	/*
	 * Truncate "unnecessary" bytes off of the most significant end of this
	 * 2's complement integer.  There should be no sequence of 9
	 * consecutive 1's or 0's at the most significant end of the
	 * integer.
	 */
	mask = 0x1FF << ((8 * (sizeof(int32) - 1)) - 1);
	/* mask is 0xFF800000 on a big-endian machine */
	while((((integer & mask) == 0) || ((integer & mask) == mask))
			&& intsize > 1){
		intsize--;
		integer <<= 8;
	}
	data = asn_build_header(data, datalength, type, intsize);
	if (data == NULL)
		return NULL;
	if (*datalength < intsize)
		return NULL;
	*datalength -= intsize;
	mask = 0xFF << (8 * (sizeof(int32) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	while(intsize--){
		*data++ = (u_char)((integer & mask) >> (8 * (sizeof(int32) - 1)));
		integer <<= 8;
	}
	return data;
}


/*
 * asn_build_unsigned_int - builds an ASN object containing an integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_unsigned_int(data, datalength, type, intp, intsize)
	u_char *data;	/* IN - pointer to start of output buffer */
	int    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	/* IN - asn type of object */
	u_long *intp;	/* IN - pointer to start of long integer */
	int    intsize;    /* IN - size of *intp */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */

	u_long integer;
	u_long mask;
	int add_null_byte = 0;

	if (intsize != sizeof (long)) {
		ERROR("not long");
		return NULL;
	}
	integer = *intp;
	mask = 0xFF << (8 * (sizeof(int32) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	if ((u_char)((integer & mask) >> (8 * (sizeof(int32) - 1))) & 0x80){
		/* if MSB is set */
		add_null_byte = 1;
		intsize++;
	}

	/* 
	 * Truncate "unnecessary" bytes off of the most significant end of
	 * this 2's complement integer.  There should be no sequence of 9
	 * consecutive 0's at the most significant end of the integer. 
	 */
	mask = 0x1FF << ((8 * (sizeof(int32) - 1)) - 1);
	/* mask is 0xFF800000 on a big-endian machine */
	while (((integer & mask) == 0) && intsize > 1){
		intsize--;
		integer <<= 8;
	}
	data = asn_build_header(data, datalength, type, intsize);
	if (data == NULL)
		return NULL;
	if (*datalength < intsize)
		return NULL;
	*datalength -= intsize;
	if (add_null_byte == 1){
		*data++ = '\0';
		intsize--;
	}
	mask = 0xFF << (8 * (sizeof(int32) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	while(intsize--){
		*data++ = (u_char)((integer & mask) >> (8 * (sizeof(int32) - 1)));
		integer <<= 8;
	}
	return data;
}


/*
 * asn_parse_string - pulls an octet string out of an ASN octet string type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the octet string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_string(data, datalength, type, string, strlength)
	u_char	    *data;	    /* IN - pointer to start of object */
	int    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    *type;	    /* OUT - asn type of object */
	u_char	    *string;	    /* IN/OUT - pointer to start of output buffer */
	int    *strlength;     /* IN/OUT - size of output buffer */
{
	/*
	 * ASN.1 octet string ::= primstring | cmpdstring
	 * primstring ::= 0x04 asnlength byte {byte}*
	 * cmpdstring ::= 0x24 asnlength string {string}*
	 */
	u_char *bufp = data;
	u_long	    asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL)
		return NULL;
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	if (asn_length > *strlength){
		ERROR("I don't support such long strings");
		return NULL;
	}
	bcopy((char *)bufp, (char *)string, (int)asn_length);
	*strlength = (int)asn_length;
	*datalength -= (int)asn_length + (bufp - data);
	return bufp + asn_length;
}


/*
 * asn_build_string - Builds an ASN octet string object containing the input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_string(data, datalength, type, string, strlength)
	u_char	    *data;	    /* IN - pointer to start of object */
	int    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	    /* IN - ASN type of string */
	u_char	    *string;	    /* IN - pointer to start of input buffer */
	int    strlength;	    /* IN - size of input buffer */
{
	/*
	 * ASN.1 octet string ::= primstring | cmpdstring
	 * primstring ::= 0x04 asnlength byte {byte}*
	 * cmpdstring ::= 0x24 asnlength string {string}*
	 * This code will never send a compound string.
	 */
	data = asn_build_header(data, datalength, type, strlength);
	if (data == NULL)
		return NULL;
	if (*datalength < strlength)
		return NULL;
	bcopy((char *)string, (char *)data, strlength);
	*datalength -= strlength;
	return data + strlength;
}


/*
 * asn_parse_header - interprets the ID and length of the current object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_header(data, datalength, type)
	u_char	    *data;	/* IN - pointer to start of object */
	int		    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    *type;	/* OUT - ASN type of object */
{
	u_char *bufp = data;
	int header_len;
	u_long asn_length;

	if (*datalength <= 0) {
		return NULL;
	}

	/* this only works on data types < 30, i.e. no extension octets */
	if (IS_EXTENSION_ID(*bufp)){
		ERROR("can't process ID >= 30");
		return NULL;
	}
	*type = *bufp;
	bufp = asn_parse_length(bufp + 1, &asn_length);
	if (bufp == NULL)
		return NULL;
	header_len = bufp - data;
	if (header_len + asn_length > *datalength){
		ERROR("asn length too long");
		return NULL;
	}
	*datalength = (int)asn_length;
	return bufp;
}

/*
 * asn_build_header - builds an ASN header for an object with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
	u_char *
asn_build_header(data, datalength, type, length)
	u_char *data;	/* IN - pointer to start of object */
	int		    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	/* IN - ASN type of object */
	int		    length;	/* IN - length of object */
{
	if (*datalength < 1)
		return NULL;
	*data++ = type;
	(*datalength)--;
	return asn_build_length(data, datalength, length);

}

/*
 * asn_build_sequence - builds an ASN header for a sequence with the ID and
 * length specified.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   in this object following the id and length.
 *
 *  This only works on data types < 30, i.e. no extension octets.
 *  The maximum length is 0xFFFF;
 *
 *  Returns a pointer to the first byte of the contents of this object.
 *  Returns NULL on any error.
 */
	u_char *
asn_build_sequence(data, datalength, type, length)
	u_char *data;	/* IN - pointer to start of object */
	int		    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	/* IN - ASN type of object */
	int		    length;	/* IN - length of object */
{
	*datalength -= 4;
	if (*datalength < 0){
		*datalength += 4;	/* fix up before punting */
		return NULL;
	}
	*data++ = type;
	*data++ = (u_char)(0x02 | ASN_LONG_LEN);
	*data++ = (u_char)((length >> 8) & 0xFF);
	*data++ = (u_char)(length & 0xFF);
	return data;
}

/*
 * asn_parse_length - interprets the length of the current object.
 *  On exit, length contains the value of this length field.
 *
 *  Returns a pointer to the first byte after this length
 *  field (aka: the start of the data field).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_length(data, length)
	u_char  *data;	/* IN - pointer to start of length field */
	u_long  *length;	/* OUT - value of length field */
{
	u_char lengthbyte = *data;

	*length = 0;
	if (lengthbyte & ASN_LONG_LEN){
		lengthbyte &= ~ASN_LONG_LEN;	/* turn MSb off */
		if (lengthbyte == 0){
			ERROR("We don't support indefinite lengths");
			return NULL;
		}
		if (lengthbyte > sizeof(long)){
			ERROR("we can't support data lengths that long");
			return NULL;
		}
		bcopy((char *)data + 1, (char *)length, (int)lengthbyte);
		/* XXX: is this useable on a 64bit platform ? */
		*length = ntohl(*length);
		*length >>= (8 * ((sizeof (*length)) - lengthbyte));
		return data + lengthbyte + 1;
	} else { /* short asnlength */
		*length = (long)lengthbyte;
		return data + 1;
	}
}

	u_char *
asn_build_length(data, datalength, length)
	u_char *data;	/* IN - pointer to start of object */
	int		    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	int    length;	/* IN - length of object */
{
	u_char    *start_data = data;

	/* no indefinite lengths sent */
	if (length < 0x80){
		if (*datalength < 1){
			ERROR("build_length");
			return NULL;
		}	    
		*data++ = (u_char)length;
	} else if (length <= 0xFF){
		if (*datalength < 2){
			ERROR("build_length");
			return NULL;
		}	    
		*data++ = (u_char)(0x01 | ASN_LONG_LEN);
		*data++ = (u_char)length;
	} else { /* 0xFF < length <= 0xFFFF */
		if (*datalength < 3){
			ERROR("build_length");
			return NULL;
		}	    
		*data++ = (u_char)(0x02 | ASN_LONG_LEN);
		*data++ = (u_char)((length >> 8) & 0xFF);
		*data++ = (u_char)(length & 0xFF);
	}
	*datalength -= (data - start_data);
	return data;

}

/*
 * asn_parse_objid - pulls an object indentifier out of an ASN object identifier type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "objid" is filled with the object identifier.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_objid(data, datalength, type, objid, objidlength)
	u_char	    *data;	    /* IN - pointer to start of object */
	int		    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    *type;	    /* OUT - ASN type of object */
	oid		    *objid;	    /* IN/OUT - pointer to start of output buffer */
	int		    *objidlength;     /* IN/OUT - number of sub-id's in objid */
{
	/*
	 * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
	 * subidentifier ::= {leadingbyte}* lastbyte
	 * leadingbyte ::= 1 7bitvalue
	 * lastbyte ::= 0 7bitvalue
	 */
	u_char *bufp = data;
	oid *oidp = objid + 1;
	u_long subidentifier;
	long   length;
	u_long	    asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL)
		return NULL;
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	*datalength -= (int)asn_length + (bufp - data);

	/* Handle invalid object identifier encodings of the form 06 00 robustly */
	if (asn_length == 0)
		objid[0] = objid[1] = 0;

	length = asn_length;
	(*objidlength)--;	/* account for expansion of first byte */
	while (length > 0 && (*objidlength)-- > 0){
		subidentifier = 0;
		do {	/* shift and add in low order 7 bits */
			subidentifier = (subidentifier << 7) + (*(u_char *)bufp & ~ASN_BIT8);
			length--;
		} while (*(u_char *)bufp++ & ASN_BIT8);	/* last byte has high bit clear */
		if (subidentifier > (u_long)MAX_SUBID){
			ERROR("subidentifier too long");
			return NULL;
		}
		*oidp++ = (oid)subidentifier;
	}

	/*
	 * The first two subidentifiers are encoded into the first component
	 * with the value (X * 40) + Y, where:
	 *	X is the value of the first subidentifier.
	 *  Y is the value of the second subidentifier.
	 */
	subidentifier = (u_long)objid[1];
	if (subidentifier == 0x2B){
		objid[0] = 1;
		objid[1] = 3;
	} else {
		objid[1] = (u_char)(subidentifier % 40);
		objid[0] = (u_char)((subidentifier - objid[1]) / 40);
	}

	*objidlength = (int)(oidp - objid);
	return bufp;
}

/*
 * asn_build_objid - Builds an ASN object identifier object containing the
 * input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_objid(data, datalength, type, objid, objidlength)
	u_char *data;	    /* IN - pointer to start of object */
	int		    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	    /* IN - ASN type of object */
	oid		    *objid;	    /* IN - pointer to start of input buffer */
	int		    objidlength;    /* IN - number of sub-id's in objid */
{
	/*
	 * ASN.1 objid ::= 0x06 asnlength subidentifier {subidentifier}*
	 * subidentifier ::= {leadingbyte}* lastbyte
	 * leadingbyte ::= 1 7bitvalue
	 * lastbyte ::= 0 7bitvalue
	 */
	u_char buf[MAX_OID_LEN];
	u_char *bp = buf;
	oid *op = objid;
	int    asnlength;
	u_long subid, mask, testmask;
	int bits, testbits;

	if (objidlength < 2){
		*bp++ = 0;
		objidlength = 0;
	} else {
		*bp++ = op[1] + (op[0] * 40);
		objidlength -= 2;
		op += 2;
	}

	while(objidlength-- > 0){
		subid = *op++;
		if (subid < 127){ /* off by one? */
			*bp++ = subid;
		} else {
			mask = 0x7F; /* handle subid == 0 case */
			bits = 0;
			/* testmask *MUST* !!!! be of an unsigned type */
			for(testmask = 0x7F, testbits = 0; testmask != 0;
					testmask <<= 7, testbits += 7){
				if (subid & testmask){	/* if any bits set */
					mask = testmask;
					bits = testbits;
				}
			}
			/* mask can't be zero here */
			for(;mask != 0x7F; mask >>= 7, bits -= 7){
				/* fix a mask that got truncated above */
				if (mask == 0x1E00000)  
					mask = 0xFE00000;
				*bp++ = (u_char)(((subid & mask) >> bits) | ASN_BIT8);
			}
			*bp++ = (u_char)(subid & mask);
		}
	}
	asnlength = bp - buf;
	data = asn_build_header(data, datalength, type, asnlength);
	if (data == NULL)
		return NULL;
	if (*datalength < asnlength)
		return NULL;
	bcopy((char *)buf, (char *)data, asnlength);
	*datalength -= asnlength;
	return data + asnlength;
}

/*
 * asn_parse_null - Interprets an ASN null type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_null(data, datalength, type)
	u_char	    *data;	    /* IN - pointer to start of object */
	int		    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    *type;	    /* OUT - ASN type of object */
{
	/*
	 * ASN.1 null ::= 0x05 0x00
	 */
	u_char   *bufp = data;
	u_long	    asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL)
		return NULL;
	if (asn_length != 0){
		ERROR("Malformed NULL");
		return NULL;
	}
	*datalength -= (bufp - data);
	return bufp + asn_length;
}


/*
 * asn_build_null - Builds an ASN null object.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_null(data, datalength, type)
	u_char	    *data;	    /* IN - pointer to start of object */
	int		    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	    /* IN - ASN type of object */
{
	/*
	 * ASN.1 null ::= 0x05 0x00
	 */
	return asn_build_header(data, datalength, type, 0);
}

/*
 * asn_parse_bitstring - pulls a bitstring out of an ASN bitstring type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  "string" is filled with the bit string.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_bitstring(data, datalength, type, string, strlength)
	u_char	    *data;	    /* IN - pointer to start of object */
	int    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    *type;	    /* OUT - asn type of object */
	u_char	    *string;	    /* IN/OUT - pointer to start of output buffer */
	int    *strlength;     /* IN/OUT - size of output buffer */
{
	/*
	 * bitstring ::= 0x03 asnlength unused {byte}*
	 */
	u_char *bufp = data;
	u_long	    asn_length;

	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL)
		return NULL;
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	if (asn_length > *strlength){
		ERROR("I don't support such long bitstrings");
		return NULL;
	}
	if (asn_length < 1){
		ERROR("Invalid bitstring");
		return NULL;
	}
	if (/** *bufp < 0 || **/ *bufp > 7){
		ERROR("Invalid bitstring");
		return NULL;
	}
	bcopy((char *)bufp, (char *)string, (int)asn_length);
	*strlength = (int)asn_length;
	*datalength -= (int)asn_length + (bufp - data);
	return bufp + asn_length;
}


/*
 * asn_build_bitstring - Builds an ASN bit string object containing the
 * input string.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the beginning of the next object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_bitstring(data, datalength, type, string, strlength)
	u_char	    *data;	    /* IN - pointer to start of object */
	int    *datalength;    /* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	    /* IN - ASN type of string */
	u_char	    *string;	    /* IN - pointer to start of input buffer */
	int    strlength;	    /* IN - size of input buffer */
{
	/*
	 * ASN.1 bit string ::= 0x03 asnlength unused {byte}*
	 */
	if (strlength < 1 || /** *string < 0 || **/ *string > 7){
		ERROR("Building invalid bitstring");
		return NULL;
	}
	data = asn_build_header(data, datalength, type, strlength);
	if (data == NULL)
		return NULL;
	if (*datalength < strlength)
		return NULL;
	bcopy((char *)string, (char *)data, strlength);
	*datalength -= strlength;
	return data + strlength;
}


/*
 * asn_parse_unsigned_int64 - pulls a 64 bit unsigned long out of an ASN int
 * type.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_parse_unsigned_int64(data, datalength, type, cp, countersize)
	u_char	    *data;	/* IN - pointer to start of object */
	int	    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char		    *type;	/* OUT - asn type of object */
	struct counter64	    *cp;	/* IN/OUT -pointer to counter struct */
	int			    countersize;/* IN - size of output buffer */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */
	u_char *bufp = data;
	u_long	    asn_length;
	u_long low = 0, high = 0;
	int intsize = 4;

	if (countersize != sizeof(struct counter64)){
		ERROR("not counter64 size");
		return NULL;
	}
	*type = *bufp++;
	bufp = asn_parse_length(bufp, &asn_length);
	if (bufp == NULL){
		ERROR("bad length");
		return NULL;
	}
	if (asn_length + (bufp - data) > *datalength){
		ERROR("overflow of message");
		return NULL;
	}
	if ((asn_length > (intsize * 2 + 1)) ||
			((asn_length == (intsize * 2) + 1) && *bufp != 0x00)){
		ERROR("I don't support such large integers");
		return NULL;
	}
	*datalength -= (int)asn_length + (bufp - data);
	if (*bufp & 0x80){
		low = -1; /* integer is negative */
		high = -1;
	}
	while(asn_length--){
		high = (high << 8) | ((low & 0xFF000000) >> 24);
		low = (low << 8) | *bufp++;
	}
	cp->low = low;
	cp->high = high;
	return bufp;
}


/*
 * asn_build_unsigned_int64 - builds an ASN object containing a 64 bit integer.
 *  On entry, datalength is input as the number of valid bytes following
 *   "data".  On exit, it is returned as the number of valid bytes
 *   following the end of this object.
 *
 *  Returns a pointer to the first byte past the end
 *   of this object (i.e. the start of the next object).
 *  Returns NULL on any error.
 */
	u_char *
asn_build_unsigned_int64(data, datalength, type, cp, countersize)
	u_char *data;	/* IN - pointer to start of output buffer */
	int    *datalength;/* IN/OUT - number of valid bytes left in buffer */
	u_char	    type;	/* IN - asn type of object */
	struct counter64 *cp;	/* IN - pointer to counter struct */
	int    countersize; /* IN - size of *intp */
{
	/*
	 * ASN.1 integer ::= 0x02 asnlength byte {byte}*
	 */

	u_int32 low, high;
	u_int32 mask, mask2;
	int add_null_byte = 0;
	int intsize;

	if (countersize != sizeof (struct counter64)) {
		ERROR("not counter64 size");
		return NULL;
	}
	intsize = 8;
	low = cp->low;
	high = cp->high;
	mask = 0xFF << (8 * (sizeof(int32) - 1));
	/* mask is 0xFF000000 on a big-endian machine */
	if ((u_char)((high & mask) >> (8 * (sizeof(int32) - 1))) & 0x80) {
		/* if MSB is set */
		add_null_byte = 1;
		intsize++;
	}
	/*
	 * Truncate "unnecessary" bytes off of the most significant end of this 2's
	 * complement integer.
	 * There should be no sequence of 9 consecutive 1's or 0's at the most
	 * significant end of the integer.
	 */
	mask2 = 0x1FF << ((8 * (sizeof(int32) - 1)) - 1);
	/* mask2 is 0xFF800000 on a big-endian machine */
	while((((high & mask2) == 0) || ((high & mask2) == mask2))
			&& intsize > 1){
		intsize--;
		high = (high << 8)
			| ((low & mask) >> (8 * (sizeof(int32) - 1)));
		low <<= 8;
	}
	data = asn_build_header(data, datalength, type, intsize);
	if (data == NULL)
		return NULL;
	if (*datalength < intsize)
		return NULL;
	*datalength -= intsize;
	if (add_null_byte == 1){
		*data++ = '\0';
		intsize--;
	}
	while(intsize--){
		*data++ = (u_char)((high & mask) >> (8 * (sizeof(int32) - 1)));
		high = (high << 8)
			| ((low & mask) >> (8 * (sizeof(int32) - 1)));
		low <<= 8;

	}
	return data;
}




struct snmp_pdu {
#ifdef HAVE_IPX
	struct sockaddr address;	/* Address of peer (ip or ipx addr) */
#else
	ipaddr address;		/* Address of peer (ip addr) */
#endif

	int	    command;		/* Type of this PDU */

	Parameters params;

	u_long  reqid;	/* Request id */
	u_long  errstat;	/* Error status */
	u_long  errindex;	/* Error index */

	/* Trap information */
	oid	    *enterprise;	/* System OID */
	int	    enterprise_length;
	ipaddr  agent_addr;		/* address of object generating trap */
	/* trap/specific type was int, now long.  
	   should not harm 32 platforms and help other: */
	long    trap_type;		/* trap type */
	long    specific_type;	/* specific type */
	u_long  time;		/* Uptime */

	struct variable_list *variables;
};


struct snmp_session {
	int	    retries;		/* Number of retries before timeout. */
	long    timeout;    	/* Number of uS until first timeout, 
							   then exponential backoff */
	char    *peername;		/* Domain name or dotted IP address or IPX 
							   address of default peer */
	u_short remote_port;	/* UDP port number of peer. */
	u_short local_port; 	/* My UDP port number, 0 for default, 
							   picked randomly */
#ifdef HAVE_IPX
	u_short remote_port_ipx;	/* IPX port number of peer. */
	u_short local_port_ipx;	/* My IPX port number, 0 for default, 
							   picked randomly */
#endif
	/* Authentication function or NULL if null authentication is used */
	int    (*authenticator)();
#if 1
	/* Function to interpret incoming data */
	int	    (*callback)  _ANSI_ARGS_(( int operation,
				struct snmp_session *session,
				int reqid,
				struct snmp_pdu *response,
				void *magic,
				... ));
#else
	int (*callback)();
#endif

	/* Pointer to data that the callback function may consider important */
	void    *callback_magic;

	int		version; 	/* SNMP version number */

	/* fields to support SNMPv1 community model */
	int	        community_len;
	u_char      *community;

	/* the private keys to use for user-based security */
	u_char   	authKey[16];
	u_char   	privKey[16];

	/* fields to support user-based security model in SNMPv2 */
	Parameters	params;

	u_char	qoS;
	u_char	agentID[16];
	u_long	agentBoots;
	int		agentTime;		/* the agentTime value */
	int		agentClock;		/* the running agentClock */
	int	     	userLen;
	u_char   	userName[32];
	int		MMS;
	int		contextLen;
	u_char	contextSelector[64];

	/* misc stuff */
	int		readView;
	int		writeView;
};

	u_char *
snmp_parse_var_op(data, var_name, var_name_len, var_val_type, var_val_len, var_val, listlength)
	u_char  *data;	    /* IN - pointer to the start of object */
	oid	    *var_name;	    /* OUT - object id of variable */
	int	    *var_name_len;  /* IN/OUT - length of variable name */
	u_char  *var_val_type;  /* OUT - type of variable (int or octet string) (one byte) */
	int	    *var_val_len;   /* OUT - length of variable */
	u_char  **var_val;	    /* OUT - pointer to ASN1 encoded value of variable */
	int	    *listlength;    /* IN/OUT - number of valid bytes left in var_op_list */
{
	u_char	    var_op_type;
	int		    var_op_len = *listlength;
	u_char	    *var_op_start = data;

	data = asn_parse_header(data, &var_op_len, &var_op_type);
	if (data == NULL){
		ERROR("snmp_parse_var_op(): 1 asn_parse_header() == NULL");
		return NULL;
	}
	if (var_op_type != (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR))
		return NULL;
	data = asn_parse_objid(data, &var_op_len, &var_op_type, var_name, var_name_len);
	if (data == NULL){
		ERROR("snmp_parse_var_op(): asn_parse_objid() == NULL");
		return NULL;
	}
	if (var_op_type != (u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID))
		return NULL;
	*var_val = data;	/* save pointer to this object */
	/* find out what type of object this is */
	data = asn_parse_header(data, &var_op_len, var_val_type);
	if (data == NULL){
		ERROR("snmp_parse_var_op(): 2 asn_parse_header() == NULL");
		return NULL;
	}
	*var_val_len = var_op_len;
	data += var_op_len;
	*listlength -= (int)(data - var_op_start);
	return data;
}

u_char *
snmp_build_var_op(data, var_name, var_name_len, var_val_type, var_val_len,
		var_val, listlength)
u_char	*data;		/* IN - pointer to the beginning of the output buffer */
oid		*var_name;	/* IN - object id of variable */
int		*var_name_len;	/* IN - length of object id */
u_char	var_val_type;	/* IN - type of variable */
int		var_val_len;	/* IN - length of variable */
u_char	*var_val;	/* IN - value of variable */
int		*listlength;	/* IN/OUT - number of valid bytes left in
						   output buffer */
{
	int		    dummyLen, headerLen;
	u_char	    *dataPtr;

	dummyLen = *listlength;
	dataPtr = data;
#if 0
	data = asn_build_sequence(data, &dummyLen,
			(u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), 0);
	if (data == NULL){
		ERROR("");
		return NULL;
	}
#endif
	data += 4;
	dummyLen -=4;
	if (dummyLen < 0)
		return NULL;

	headerLen = data - dataPtr;
	*listlength -= headerLen;
	data = asn_build_objid(data, listlength,
			(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID),
			var_name, *var_name_len);
	if (data == NULL){
		ERROR("");
		return NULL;
	}
	switch(var_val_type){
		case ASN_INTEGER:
			data = asn_build_int(data, listlength, var_val_type,
					(long *)var_val, var_val_len);
			break;
		case GAUGE:
		case COUNTER:
		case TIMETICKS:
		case UINTEGER:
			data = asn_build_unsigned_int(data, listlength, var_val_type,
					(u_long *)var_val, var_val_len);
			break;
		case COUNTER64:
			data = asn_build_unsigned_int64(data, listlength, var_val_type,
					(struct counter64 *)var_val,
					var_val_len);
			break;
		case ASN_OCTET_STR:
		case IPADDRESS:
		case OPAQUE:
		case NSAP:
			data = asn_build_string(data, listlength, var_val_type,
					var_val, var_val_len);
			break;
		case ASN_OBJECT_ID:
			data = asn_build_objid(data, listlength, var_val_type,
					(oid *)var_val, var_val_len / sizeof(oid));
			break;
		case ASN_NULL:
			data = asn_build_null(data, listlength, var_val_type);
			break;
		case ASN_BIT_STR:
			data = asn_build_bitstring(data, listlength, var_val_type,
					var_val, var_val_len);
			break;
		case SNMP_NOSUCHOBJECT:
		case SNMP_NOSUCHINSTANCE:
		case SNMP_ENDOFMIBVIEW:
			data = asn_build_null(data, listlength, var_val_type);
			break;
		default:
			ERROR("wrong type");
			return NULL;
	}
	if (data == NULL){
		ERROR("");
		return NULL;
	}
	dummyLen = (data - dataPtr) - headerLen;

	asn_build_sequence(dataPtr, &dummyLen,
			(u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), dummyLen);
	return data;
}

/*
   static int
   snmp_parse(session, pdu, data, length)
   struct snmp_session *session; */
	static int
snmp_parse( pdu, data, length)
	struct snmp_pdu *pdu;
	u_char  *data;
	int	    length;
{
	u_char  msg_type;
	u_char  type;
	u_char  *var_val;
	long    version;
	int	    len, four;
	u_char community[256];
	int community_length = 256;
	struct variable_list *vp = 0;
	oid	    objid[MAX_NAME_LEN], *op;
	u_char  *origdata = data;
	int      origlength = length;
	int      ret = 0;
	u_char  *save_data;

	/* authenticates message and returns length if valid */
	data = snmp_auth_parse(data, &length, community, &community_length, &version);
	if (data == NULL)
		return -1;

	if( version != SNMP_VERSION_1 && version != SNMP_VERSION_2C && version != SNMP_VERSION_2 ) {
		fprintf(stderr, "Wrong version: %ld\n", version);
		return -1;
	}

	save_data = data;

	data = asn_parse_header(data, &length, &msg_type);
	if (data == NULL)
		return -1;
	pdu->command = msg_type;

	/* if( session->authenticator ) {
	   ret = session->authenticator( origdata, origlength, save_data - community_length, 
	   community_length, session, pdu ); */
	if( ret < 0 ) return ret;
}

if (pdu->command != TRP_REQ_MSG){
	data = asn_parse_int(data, &length, &type, (long *)&pdu->reqid, sizeof(pdu->reqid));
	if (data == NULL)
		return -1;
	data = asn_parse_int(data, &length, &type, (long *)&pdu->errstat, sizeof(pdu->errstat));
	if (data == NULL)
		return -1;
	data = asn_parse_int(data, &length, &type, (long *)&pdu->errindex, sizeof(pdu->errindex));
	if (data == NULL)
		return -1;
} else {
	pdu->enterprise_length = MAX_NAME_LEN;
	data = asn_parse_objid(data, &length, &type, objid, &pdu->enterprise_length);
	if (data == NULL)
		return -1;
	pdu->enterprise = (oid *)calloc(1, pdu->enterprise_length * sizeof(oid));
	bcopy((char *)objid, (char *)pdu->enterprise, pdu->enterprise_length * sizeof(oid));

	four = 4;
	data = asn_parse_string(data, &length, &type, (u_char *)&pdu->agent_addr.sin_addr.s_addr, &four);
	if (data == NULL)
		return -1;
	data = asn_parse_int(data, &length, &type, (long *)&pdu->trap_type, sizeof(pdu->trap_type));
	if (data == NULL)
		return -1;
	data = asn_parse_int(data, &length, &type, (long *)&pdu->specific_type, sizeof(pdu->specific_type));
	if (data == NULL)
		return -1;
	data = asn_parse_int(data, &length, &type, (long *)&pdu->time, sizeof(pdu->time));
	if (data == NULL)
		return -1;
}
data = asn_parse_header(data, &length, &type);
if (data == NULL)
	return -1;
if (type != (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR))
	return -1;
	while((int)length > 0){
		if (pdu->variables == NULL){
			pdu->variables = vp = (struct variable_list *)calloc(1, sizeof(struct variable_list));
		} else {
			vp->next_variable = (struct variable_list *)calloc(1, sizeof(struct variable_list));
			vp = vp->next_variable;
		}
		vp->next_variable = NULL;
		vp->val.string = NULL;
		vp->name = NULL;
		vp->name_length = MAX_NAME_LEN;
		data = snmp_parse_var_op(data, objid, &vp->name_length, &vp->type, &vp->val_len, &var_val, (int *)&length);
		if (data == NULL)
			return -1;
		op = (oid *)calloc(1, (unsigned)vp->name_length * sizeof(oid));
		bcopy((char *)objid, (char *)op, vp->name_length * sizeof(oid));
		vp->name = op;

		len = PACKET_LENGTH;
		switch((short)vp->type){
			case ASN_INTEGER:
				vp->val.integer = (long *)calloc(1, sizeof(long));
				vp->val_len = sizeof(long);
				asn_parse_int(var_val, &len, &vp->type, (long *)vp->val.integer, sizeof(vp->val.integer));
				break;
			case COUNTER:
			case GAUGE:
			case TIMETICKS:
			case UINTEGER:
				vp->val.integer = (long *)calloc(1, sizeof(unsigned long));
				vp->val_len = sizeof(unsigned long);
				asn_parse_unsigned_int(var_val, &len, &vp->type, (unsigned long *)vp->val.integer, sizeof(vp->val.integer));
				break;
			case COUNTER64:
				vp->val.counter64 = (struct counter64 *)calloc(1,  sizeof(struct counter64) );
				vp->val_len = sizeof(struct counter64);
				asn_parse_unsigned_int64(var_val, &len, &vp->type,
						(struct counter64 *)vp->val.counter64,
						sizeof(*vp->val.counter64));
				break;
			case ASN_OCTET_STR:
			case IPADDRESS:
			case OPAQUE:
			case NSAP:
				vp->val.string = (u_char *) calloc (1, 
						(unsigned)vp->val_len + 1);
				asn_parse_string(var_val, &len, &vp->type, vp->val.string,
						&vp->val_len);
				break;
			case ASN_OBJECT_ID:
				vp->val_len = MAX_NAME_LEN;
				asn_parse_objid(var_val, &len, &vp->type, objid, &vp->val_len);
				vp->val_len *= sizeof(oid);
				vp->val.objid = (oid *)calloc(1, (unsigned)vp->val_len);
				bcopy((char *)objid, (char *)vp->val.objid, vp->val_len);
				break;
			case SNMP_NOSUCHOBJECT:
			case SNMP_NOSUCHINSTANCE:
			case SNMP_ENDOFMIBVIEW:
			case ASN_NULL:
				break;
			default:
				fprintf(stderr, "bad type returned (%x)\n", vp->type);
				break;
		}
	}
return ret;
}



/*
 * Takes a session and a pdu and serializes the ASN PDU into the area
 * pointed to by packet.  out_length is the size of the data area available.
 * Returns the length of the completed packet in out_length.  If any errors
 * occur, -1 is returned.  If all goes well, 0 is returned.
 */
	int
snmp_build( pdu, packet, out_length, is_agent)
	//   struct snmp_session	*session;
	struct snmp_pdu	*pdu;
	u_char	*packet;
	int			*out_length;
	int			 is_agent;
{
	u_char  buf[PACKET_LENGTH];
	u_char  *cp;
	struct variable_list *vp;
	int	    length;
	int	    totallength;

	length = *out_length;
	cp = packet;
	for(vp = pdu->variables; vp; vp = vp->next_variable){
		cp = snmp_build_var_op(cp, vp->name, &vp->name_length, vp->type, vp->val_len, (u_char *)vp->val.string, &length);
		if (cp == NULL)
			return -1;
	}
	totallength = cp - packet;

	length = PACKET_LENGTH;
	cp = asn_build_header(buf, &length, (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), totallength);
	if (cp == NULL)
		return -1;
	bcopy((char *)packet, (char *)cp, totallength);
	totallength += cp - buf;

	length = *out_length;
	if (pdu->command != TRP_REQ_MSG){
		/* request id */
		cp = asn_build_int(packet, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
				(long *)&pdu->reqid, sizeof(pdu->reqid));
		if (cp == NULL)
			return -1;
		/* error status */
		cp = asn_build_int(cp, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
				(long *)&pdu->errstat, sizeof(pdu->errstat));
		if (cp == NULL)
			return -1;
		/* error index */
		cp = asn_build_int(cp, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
				(long *)&pdu->errindex, sizeof(pdu->errindex));
		if (cp == NULL)
			return -1;
	} else {	/* this is a trap message */
		/* enterprise */
		cp = asn_build_objid(packet, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OBJECT_ID),
				(oid *)pdu->enterprise, pdu->enterprise_length);
		if (cp == NULL)
			return -1;
		/* agent-addr */
		cp = asn_build_string(cp, &length, (u_char)IPADDRESS,
				(u_char *)&pdu->agent_addr.sin_addr.s_addr, sizeof(pdu->agent_addr.sin_addr.s_addr));
		if (cp == NULL)
			return -1;
		/* generic trap */
		cp = asn_build_int(cp, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
				(long *)&pdu->trap_type, sizeof(pdu->trap_type));
		if (cp == NULL)
			return -1;
		/* specific trap */
		cp = asn_build_int(cp, &length,
				(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
				(long *)&pdu->specific_type, sizeof(pdu->specific_type));
		if (cp == NULL)
			return -1;
		/* timestamp  */
		cp = asn_build_int(cp, &length, (u_char)TIMETICKS,
				(long *)&pdu->time, sizeof(pdu->time));
		if (cp == NULL)
			return -1;
	}
	if (length < totallength)
		return -1;
	bcopy((char *)buf, (char *)cp, totallength);
	totallength += cp - packet;

	length = PACKET_LENGTH;
	cp = asn_build_header(buf, &length, (u_char)pdu->command, totallength);
	if (cp == NULL)
		return -1;
	if (length < totallength)
		return -1;
	bcopy((char *)packet, (char *)cp, totallength);
	totallength += cp - buf;

	length = *out_length;


	//cp = snmp_auth_build( packet, &length, session is_agent, totallength );

	cp = snmp_auth_build( packet, &length, is_agent, totallength );
	if (cp == NULL)
		return -1;
	if ((*out_length - (cp - packet)) < totallength)
		return -1;
	bcopy((char *)buf, (char *)cp, totallength);
	totallength += cp - packet;
	*out_length = totallength;
	/*
	   if( session->qoS & USEC_QOS_AUTH )
	   md5Digest( packet, totallength, cp - (session->contextLen + 16),
	   cp - (session->contextLen + 16) );
	   */
	return 0;
}
	u_char *
snmp_auth_parse(data, length, sid, slen, version)
	u_char	    *data;
	int		    *length;
	u_char	    *sid;
	int		    *slen;
	long	    *version;
{
	u_char    type;

	data = asn_parse_header(data, length, &type);
	if (data == NULL){
		ERROR("bad header");
		return NULL;
	}
	if (type != (ASN_SEQUENCE | ASN_CONSTRUCTOR)){
		ERROR("wrong auth header type");
		return NULL;
	}
	data = asn_parse_int(data, length, &type, version, sizeof(*version));
	if (data == NULL){
		ERROR("bad parse of version");
		return NULL;
	}
	data = asn_parse_string(data, length, &type, sid, slen);
	if (data == NULL){
		ERROR("bad parse of community");
		return NULL;
	}

	if( *version == SNMP_VERSION_1 ) sid[*slen] = '\0';
	return (u_char *)data;
}

// Carefully remove session from this function

//u_char *snmp_auth_build( data, length, session, is_agent, messagelen )
u_char *snmp_auth_build( data, length, is_agent, messagelen )
	u_char	            *data;
	int		    *length;
	//struct snmp_session *session;
	int		     is_agent;
	int		     messagelen;
{
	u_char *params;
	int     plen;

	if( session->version == SNMP_VERSION_2 ) {
		u_char buff[138]; /* max len of param string */
		u_char *pp = buff;
		u_long tmp;

		*pp++ = SNMP_USEC_MODEL;	/* usec model */

		*pp++ = session->qoS;

		memcpy( pp, session->agentID, 12 );
		pp += 12;

		if( is_agent || session->qoS & USEC_QOS_AUTH ) {
			tmp = session->agentBoots;
			*pp++ = (tmp >> 24) & 0xff;
			*pp++ = (tmp >> 16) & 0xff;
			*pp++ = (tmp >> 8) & 0xff;
			*pp++ = tmp & 0xff;
		} else {
			memset( pp, 0, 4 );
			pp += 4;
		}

		if( is_agent || session->qoS & USEC_QOS_AUTH ) {
			tmp = session->agentClock + time(NULL);
			*pp++ = (tmp >> 24) & 0xff;
			*pp++ = (tmp >> 16) & 0xff;
			*pp++ = (tmp >> 8) & 0xff;
			*pp++ = tmp & 0xff;
		} else {
			memset( pp, 0, 4 );
			pp += 4;
		}

		tmp = session->MMS;
		*pp++ = (tmp >> 8) & 0xff;
		*pp++ = tmp & 0xff;

		*pp++ = session->userLen;
		memcpy( pp, session->userName, session->userLen );
		pp += session->userLen;

		if( session->qoS & USEC_QOS_AUTH ) {
			*pp++ = 16;
			memcpy( pp, session->authKey, 16 );
			pp += 16;
		} else {
			*pp++ = 0;
		}

		memcpy( pp, session->contextSelector, session->contextLen );
		pp += session->contextLen;

		params = buff;
		plen = pp - buff;
	} else {
		params = session->community;
		plen = session->community_len;
	}

	data = asn_build_sequence(data, length, (u_char)(ASN_SEQUENCE | ASN_CONSTRUCTOR), 
			messagelen + plen + 5);
	if (data == NULL){
		ERROR("buildheader");
		return NULL;
	}
	data = asn_build_int(data, length,
			(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_INTEGER),
			(long *)&session->version, sizeof(session->version));
	if (data == NULL){
		ERROR("buildint");
		return NULL;
	}

	data = asn_build_string(data, length,
			(u_char)(ASN_UNIVERSAL | ASN_PRIMITIVE | ASN_OCTET_STR), params, plen );
	if (data == NULL){
		ERROR("buildstring");
		return NULL;
	}

	return (u_char *)data;
}

main ()
{ 
	//make a packet 
	//snmp_build the packet
	//snmp parse the packet
	//display the contents of the parsed string
}

