package client_header;

use strict;

sub errmsg {
    my $err = shift;
    my $mesg;

    $mesg = "Error: Value too big" if ( $err == 1 );
    $mesg = "Error: No Such Name"  if ( $err == 2 );
    $mesg = "Error: Bad Value"     if ( $err == 3 );
    $mesg = "Error: Read Only"     if ( $err == 4 );
    $mesg = "Error: Generic Error" if ( $err == 5 );
    $mesg = "EndOfMIB"             if ( $err == 6 );

    $mesg = $mesg . "\n";

    print "$mesg";
}

sub TcpStatePopulator {

    #figures out if oid is TcpConnState, and if so, populates the $val
    #with the textual meaning of the returned integer state
    my $textual_oid = shift;
    my $value       = shift;
    my $textual_value;

    if ( $textual_oid =~ m/tcpConnState/ ) {
        $textual_value = "5 (established)" if ( $value == 5 );
        $textual_value = "3 (syn_sent)"    if ( $value == 3 );
        $textual_value = "4 (syn_recv)"    if ( $value == 4 );
        $textual_value = "6 (fin_wait_1)"  if ( $value == 6 );
        $textual_value = "7 (fin_wait_2)"  if ( $value == 7 );
        $textual_value = "11 (time_wait)"  if ( $value == 11 );
        $textual_value = "1 (close)"       if ( $value == 1 );
        $textual_value = "8 (close_wait)"  if ( $value == 8 );
        $textual_value = "9 (last_ack)"    if ( $value == 9 );
        $textual_value = "2 (listen)"      if ( $value == 2 );
        $textual_value = "10 (closing)"    if ( $value == 10 );
        $textual_value = "12 (delete_tcb)" if ( $value == 12 );
    }
    else {
        $textual_value = $value;
    }

    return $textual_value;
}

sub OIDResolver {

    #find out if OID entered is name or numerical, and either way, the output is
    #a numerical OID, and an error status
    my $requested_oid = shift;

    #print("[OidResolver: oid:$requested_oid\n");
    my ( $err, $verified_oid );

    if ( $requested_oid =~ m/^[A-Za-z]+\.?.*/ ) {
        ( $err, $verified_oid ) = Name2Number($requested_oid);
    }
    elsif ( $requested_oid =~ m/^\.?(([0-9]+[.]?)+)$/ ) {
        my $temp_oid = $1;
        $temp_oid = "\." . "$temp_oid";    #put the "." in the BGNG
        chop $temp_oid if ( $temp_oid =~ m/(.*)\.$/ );    #remove trailing "."
        $verified_oid = $temp_oid;
        $err          = 0;
    }
    else {
        $err          = 5;
        $verified_oid = $requested_oid;
    }
    return ( $err, $verified_oid );
}

sub Number2Name {

    #change incoming numerical response to text based
    #if name not found then return whatever was sent back
    my $received_oid = shift;

    #print("[Number2Name]: recvd oid is:$received_oid\n");
    my $name_res = 'oid_resolution.txt';
    my ( $name, $err, $numeric, $textual, $rest_of_it );
    open( NAME_RES, $name_res );
    while (<NAME_RES>) {
        if (m/^([.0-9]+)\s+([a-zA-Z]+)$/) {
            $numeric = $1;
            $textual = $2;

            #print("[Debugging]: numeric: $numeric\n");
            #if ($received_oid =~ m/^"$numeric"(\..*)\s+([A-Za-z]+)$/)
            if ( $received_oid =~ m/^$numeric(\..*)/ ) {
                $rest_of_it = $1;
                $name       = "$textual" . "$rest_of_it";
                $err        = 0;

                #print("[Number1Name]: corresponding name:$name\n");
                last;
            }
            else {
                $name = $received_oid;
                $err  = 2;
            }
        }
    }
    close(NAME_RES);
    return ( $err, $name );
}

sub Name2Number {
    my $name_oid = shift;

    #print("[Name2Number] nameoid is:$name_oid\n");
    my ( $error, $numeric, $complete_numeric );

    my $name_res = 'oid_resolution.txt';
    open( NAME_RES, $name_res );

    #to match just system/ip/icmp/tcp:
    if ( $name_oid =~ m/^([a-zA-Z]+)$/ ) {
        my $first = $1;
        while (<NAME_RES>) {
            if (m/^([.0-9]+)\s+$first$/) {
                $numeric = $1;
                $error   = 0;
                last;
            }
            else {
                $error = 5;
            }
        }
        close(NAME_RES);
        return ( $error, $numeric );
    }
    if ( $name_oid =~ m/^([a-zA-Z]+)\.(([0-9]+[.]?)+)$/ ) {
        my $first  = $1;
        my $second = $2;

        #print("first:$first, second:$second\n");
        while (<NAME_RES>) {
            if (m/^([.0-9]+)\s+$first$/) {
                $numeric = $1;
                $error   = 0;
                last;
            }
            else {
                $error = 5;
            }
        }

        if ( $second =~ m/(.*)\.$/ ) {
            chop $second;
        }

        if ( $error == 0 ) {
            $complete_numeric = "$numeric" . ".$second";
        }
        else {
            $complete_numeric = $name_oid;
        }

    }
    else {

        #print("nopes!!\n");
        $error = 5;
    }

    close(NAME_RES);
    return ( $error, $complete_numeric );
}

1;
__END__
