#!/usr/bin/perl -w
# $	snmpSet client, for gmov-snmp snmp agent 4.20	$
use Socket;
use client_header;
use strict;

$|++;
if ( scalar(@ARGV) != 4 ) {
    print("snmpSet: [gmov-snmp v0.1]\n");
    print("usage: snmpset.pl <host> <commstring> <oid> <setval>\n");
    print("<host>		: the snmp agent you want to poll\n");
    print("<commstring>	: the community string you want to use\n");
    print("<oid>		: the OID you want to set, can be textual also\n");
    print("<setval>	: the value you want to set.\n");
    print("Use \"\" around multiworded values\n");
    print("e.g. snmpset machine.com public sysContact.0 \"Alan Greenspan\"\n");
    exit;
}
my $protocol_name = 'udp';
my $ipee          = shift;
my $commstring    = shift;
my $oid           = shift;
my $set_val       = shift;
$set_val =~ s/ /+_+/g;    # a hook to handle space
my $port            = '161';
my $sending_port    = '12347';
my $retries         = '5';
my $thatmachines_ip = inet_aton($ipee);
my $host_name       = `uname -n`;
chomp $host_name;
my $local_ip = `grep $host_name /etc/hosts | awk '{print \$1}'`;
chomp $local_ip;

my $thismachines_ip      = inet_aton($host_name);
my $local_packed_address = sockaddr_in( $sending_port, $thismachines_ip );
my $protocol_numer       = getprotobyname($protocol_name);
my $request_id           = rand(64000);
$request_id =~ s/^(\d+)\..*/$1/;

my ( $oid_err, $verified_oid, $query_response, $query, $i, $rin, $active_socket,
    $err );

#now for creating a packed address for the other end:
my $remote_packed_address = sockaddr_in( $port, $thatmachines_ip );

#now to send request over

( $oid_err, $verified_oid ) = client_header::OIDResolver($oid);

if ( $oid_err == 0 ) {
    $query = "3 $commstring $verified_oid $set_val $request_id";
}
else {
    print("Error: Bad OID\n");
    exit;
}

#print("[snmpSet client] sending query to $ipee\:$port\n");
socket( SOCK_SENDER, PF_INET, SOCK_DGRAM, $protocol_numer );
bind( SOCK_SENDER, $local_packed_address );
if ( !( send( SOCK_SENDER, $query, 0, $remote_packed_address ) ) ) {
    print("could not send\n");
}

for ( $i = 0 ; $i < $retries ; $i++ ) {
    print("*");
    $rin = '';
    vec( $rin, fileno(SOCK_SENDER), 1 ) = 1;
    $active_socket = 0;
    $active_socket = select( $rin, undef, undef, 2 );
    if ($active_socket) {

        #print("match! active_sock $active_socket\n");
        last;
    }
    else {

        #print("nomatch active_sock $active_socket, resending\n");
        send( SOCK_SENDER, $query, 0, $remote_packed_address );
    }
}

if ($active_socket) {
    die "$!"
      if ( !( recv( SOCK_SENDER, $query_response, 1024, 0 ) ) );

    ($err) = ResponseParser($query_response);
    if ( $err != 0 ) {
        client_header::errmsg($err);
        exit;
    }
}

else {
    print("timeout\n");
    exit;
}

#----end of main ----#

sub ResponseParser {
    my $qr = shift;
    my ( $named_oid, $tb, $err, $oid, $val, $req_id, $next_query );
    if ( $qr =~ m/^(\d+)\s+(\S+)\s+(.*)\s+(\d+)$/ ) {
        $err    = $1;
        $oid    = $2;
        $val    = $3;
        $req_id = $4;

        return ($err) if ( $err != 0 );
        $val =~ s/\+\_\+/ /g;

        ( $err, $named_oid ) = client_header::Number2Name($oid);
        print("${named_oid}: $val\n") if ( $err == 0 );

    }
    return $err;
}
