#!/usr/bin/perl -w
# snmpgetnext implementation for gmov-snmp
use Socket;
use client_header;
use strict;
$|++;

if ( scalar(@ARGV) != 3 ) {
    print("snmpWalk [gmov-snmp v0.1]\n");
    print("usage: snmpwalk.pl <host> <commstring> <oid>\n");
    print("<host> 		: the snmp agent you want to poll\n");
    print("<commstring>	: the community string you want to use\n");
    print("<oid>		: the OID branch you want to poll, can be textual also\n");
    exit;
}

sub RandReqId() {
    my $id = rand(64000);
    $id =~ s/^(\d+)\..*$/$1/;
    return $id;
}

#------main() -------#
my $protocol_name   = 'udp';
my $ipee            = shift;
my $commstring      = shift;
my $oid             = shift;
my $port            = '161';
my $sending_port    = '12345';
my $retries         = '5';
my $thatmachines_ip = inet_aton($ipee);
my $host_name       = `uname -n`;
chomp $host_name;
my $local_ip = `grep $host_name /etc/hosts | awk '{print \$1}'`;
chomp $local_ip;
my $thismachines_ip      = inet_aton($local_ip);
my $local_packed_address = sockaddr_in( $sending_port, $thismachines_ip );
my $protocol_number      = getprotobyname($protocol_name);
my $request_id           = RandReqId ();
my $error                = 0;
my (
    $oid_err,       $verified_oid,   $test_branch,
    $query,         $rin,            $i,
    $active_socket, $query_response, $value
);
my $original_oid = $oid;
my $header       = 0;
my $next_query;

if ( $original_oid eq "system" ) {
    $original_oid = ".1.3.6.1.2.1.1";
}
elsif ( $original_oid eq "ip" ) {
    $original_oid = ".1.3.6.1.2.1.4";
}
elsif ( $original_oid eq "icmp" ) {
    $original_oid = ".1.3.6.1.2.1.5";
}
elsif ( $original_oid eq "tcp" ) {
    $original_oid = ".1.3.6.1.2.1.6";
}
elsif ( ( !( $original_oid =~ m/\./ ) ) || ( $original_oid =~ m/\.$/ ) ) {
    $original_oid = ".1.0";
}

my @list = split ( /\./, $original_oid );
my $last_pos = scalar(@list) - 1;

$original_oid =~ s/.*(\d+)$/$1/;
$test_branch = $original_oid;

#print("orig $original_oid, test_b $test_branch, last_pos $last_pos\n");

#now for creating a packed address for the other end:
my $remote_packed_address = sockaddr_in( $port, $thatmachines_ip );

#now to send request over

( $oid_err, $verified_oid ) = client_header::OIDResolver($oid);

if ( $oid_err == 0 ) {
    $query = "1 $commstring $verified_oid -1 $request_id";
}
else {
    print("Error: Bad OID\n");
    exit;
}

$next_query = $query;

#print("query sent is $query\n");

#print("test br: $test_branch\n");
#while ( $error == 0 ) {
while ( $test_branch == $original_oid ) {

    if ( !($header) ) {
        print("[snmpwalk client] Sending queries to $ipee\:$port\n");
        $header++;
    }

    $query = $next_query;

    #print("Query Just about to be sent is: $query\n");
    select( undef, undef, undef, .1 );

    #print("[WALK: Request]: $query\n");
    socket( SOCK_SENDER, PF_INET, SOCK_DGRAM, $protocol_number );
    bind( SOCK_SENDER, $local_packed_address );
    if ( !( send( SOCK_SENDER, $query, 0, $remote_packed_address ) ) ) {
        print("could not send\n");
    }
    else {

        #print("QuerySent: $query\n");
    }

    for ( $i = 1 ; $i < $retries ; $i++ ) {
        print("*");
        $rin = '';
        vec( $rin, fileno(SOCK_SENDER), 1 ) = 1;
        $active_socket = 0;
        $active_socket = select( $rin, undef, undef, 2 );
        if ($active_socket) {

            #print("act $active_socket\n");
            #print("match! active_sock $active_socket\n");
            last;
        }
        else {

            #print("nomatch active_sock $active_socket, resending\n");
            send( SOCK_SENDER, $query, 0, $remote_packed_address );
        }
    }

    if ($active_socket) {
        die "$!"
          if ( !( recv( SOCK_SENDER, $query_response, 1024, 0 ) ) );

        #print("[WALK: Response]: $query_response\n");
        ( $error, $value, $test_branch, $next_query ) =
          ResponseParser( $query_response, $last_pos, $original_oid );

        if ( $error != 0 ) {
            client_header::errmsg($error);
            exit;
        }

    }

    else {
        print("*Timed out\n");
        exit;
    }
}

#
#end of main

sub ResponseParser {
    my $qr = shift;
    my $lp = shift;
    my $oo = shift;
    my ( $named_oid, $tb, @oid, $err, $oid, $val, $req_id, $next_query );
    if ( $qr =~ m/^(\d+)\s+(\S+)\s+(.*)\s+(\d+)$/ ) {
        $err    = $1;
        $oid    = $2;
        $val    = $3;
        $req_id = $4;
        @oid    = split ( /\./, $oid );

        #print("last_pos:$lp responseParser:oid:@oid\n");
        $tb = $oid[$lp];

        #print("test_branch: $tb\n");

        return ($err) if ( $err != 0 );
        $val =~ s/\+\_\+/ /g;

        #print("${oid}: $val \($req_id\)\n") if (($tb == $oo) && ($err == 0));

        ( $err, $named_oid ) = client_header::Number2Name($oid);
        ($val) = client_header::TcpStatePopulator( $named_oid, $val );

        print("${named_oid}: $val\n") if ( ( $tb == $oo ) && ( $err == 0 ) );

        #$print("[ResponseParser] test br_inside: $test_branch\n");

        $req_id     = RandReqId ();
        $next_query = "1 $commstring $oid -1 $req_id";
    }
    return $err, $val, $tb, $next_query;
}

