/* file handling IP branch of snmp tree */
#include <string.h>
#include <stdio.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>

#include "lib/oids_supported.h"

ip_table *ipNetToMediaTable_LL;

void NullNextOid(oid *next_oid)	{
	/*this function initializes all values in the next_oid array*/
	int i;
	for (i=6;i<20;i++)	
		next_oid[i] = -1;
	/* making all -1 */
}

void InitNextOid(oid *next_oid)	{
	/*this function ensures  the constt stuff in oid */
	/*the first 6 values are always .1.3.6.1.2.1*/
	next_oid[0] = 1; next_oid[1] = 3;next_oid[2] = 6; next_oid[3] = 1;
	next_oid[4] = 2; next_oid[5] = 1; 
}

int IpOidMatch (char *line,const oid *req_oid, int *mac_addr)
{
	//this fn is to scan the line, and
	//match the leaf portion of the IpNetToMediaAddress
	//ie. the A.B.C.D in ipNetToMediaIfIndex.X.A.B.C.D
	//returns 1 if no match, and 0 if matched.
	//also mac_addr is populated if 0 is returned

	int i1, i2, i3, i4, m1, m2, m3, m4, m5, m6;
	char ifname[20];
	int mcaddr[6];

	if (11 == sscanf (line, "%d.%d.%d.%d 0x%*x 0x%*x %x:%x:%x:%x:%x:%x %*[^ ] %20s\n",
				&i1, &i2, &i3, &i4,
				&m1, &m2, &m3, &m4, &m5, &m6, ifname))
		//keeping ifname for future growth
	{
		if (((*(req_oid+11)) == i1) && ((*(req_oid+12)) == i2)
				&& ((*(req_oid+13)) == i3) && ((*(req_oid+14)) == i4))
			//IP matched!!
		{
			mcaddr[0] = m1; mcaddr[1] = m2; mcaddr[2] = m3;
			mcaddr[3] = m4; mcaddr[4] = m5; mcaddr[5] = m6;
			memcpy(mac_addr, mcaddr, sizeof(mcaddr));
			return 0;
		} 
	} 
	return 1;
}


void IpFn (short int req_type,const oid *req_oid,const int oid_length, char *data, 
		oid *next_oid, char *set_val)
{
	char ip_addr_c[15];
	struct arpreq req;
	struct sockaddr_in *sin;
	struct hostent *serv_hostent;
	int sockfd = 0;
	/*printf("[IpFn] reqoid rcvd: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
	  req_oid[0], req_oid[1], req_oid[2], req_oid[3],
	  req_oid[4], req_oid[5], req_oid[6], req_oid[7],
	  req_oid[8], req_oid[9], req_oid[10], req_oid[11],
	  req_oid[12], req_oid[13], req_oid[14], req_oid[15]); */
	if (req_type == 0) //snmpget
	{
		//snmpget for IpFn always gets right length of query from
		//query sanity checker: thus no need to check size
		char line[128];
		int mac_addr[6];
		int iostatus = 1;
		FILE *in = fopen ("/proc/net/arp", "r");

		bzero(data, sizeof(data));

		NullNextOid(next_oid);

		OidCopier(next_oid, req_oid, oid_length);/*return the same oid in next_oid*/

		//printf("oid length %d\n", oid_length);

		/*note: this function is defined in icmpfn.c */
		while ((line == fgets(line, sizeof(line), in)) &&
				(iostatus = IpOidMatch(line, req_oid, mac_addr)))
		{
			// still havent found what i'm looking for -u2
			continue;
		}

		//printf("iostatus = %d\n",iostatus);

		if (!iostatus)	{

			switch (*(req_oid+9)) {

				case 1: //ipNetToMediaIfIndex
					//printf("ipfn: ifIndex matched!");
					sprintf(data, "%d", 1);
					fclose (in);
					break;

				case 2: //ipNetToMediaPhysAddress
					sprintf(data, "%x:%x:%x:%x:%x:%x", 
							mac_addr[0], mac_addr[1], mac_addr[2], 
							mac_addr[3], mac_addr[4], mac_addr[5]);
					fclose (in);
					break;

				case 3: //ipNetToMediaNetAddress
					sprintf(data, "%d.%d.%d.%d", 
							(*(req_oid+11)), (*(req_oid+12)), 
							(*(req_oid+13)), (*(req_oid+14)));
					fclose (in);
					break;

				case 4: //ipNetToMediaType
					sprintf(data, "%d", 1);
					fclose (in);
					break;

				default:
					//error: oid not supported\n");
					sprintf(data, "%d", -1);
					fclose (in);
					break;

			}

		} else {
			// if iostatus != 0, the leaf didnt match
			// according to IpOidMatch(), thus..
			sprintf(data, "%d", -1);
			fclose (in);
			return;
		}

	} /*end if req_type ==0 */

	/* -------SNMPGETNEXT CODING FOR IP --------*/

	else if (req_type == 1) //snmpgetnext
	{

		int i1, i2, i3, i4, m1, m2, m3, m4, m5, m6;
		char ifname[20];
		int mcaddr[6];
		char line[128];
		int iostatus = 1;
		int mac_addr[6];


		FILE *in = fopen ("/proc/net/arp", "r");
		bzero(data, sizeof(data));

		// note: we might get an OID of variable length. from
		// .1.3.6.1.2.1.4.22.1.[1-4] [10 levels]
		// .1.3.6.1.2.1.4.22.1.[1-4].X.A.B.C.D [15 levels]
		//
		if (oid_length < 60) /*remember u_long = 4bytes, and oid_length = 15*4bytes*/
		{
			//this means that the oid QSN has given us  till:
			//.1.3.6.1.2.1.4.22.1.[1-4]
			// thus we'll grab the first line in /proc/net/arp,
			// and filter out the request once we have it.
			while ((line == fgets(line, sizeof(line), in)) &&
					(11 != sscanf (line, "%d.%d.%d.%d 0x%*x 0x%*x %x:%x:%x:%x:%x:%x %*[^ ] %20s\n",
								   &i1, &i2, &i3, &i4,
								   &m1, &m2, &m3, &m4, &m5, &m6, ifname)))
			{
				continue;
			}

			mcaddr[0] = m1; mcaddr[1] = m2; mcaddr[2] = m3;
			mcaddr[3] = m4; mcaddr[4] = m5; mcaddr[5] = m6;
			memcpy(mac_addr, mcaddr, sizeof(mcaddr));
			/*first, let's copy over whatever the req_oid is over our next_oid[20]
			 * we will build the oid after this */

			NullNextOid(next_oid);
			OidCopier(next_oid, req_oid, oid_length);

			InitNextOid(next_oid); /*ensuring the first 6 constt values are there*/
			/*you never know what the client sent for a get next request! */

			/*now plastering everything else (but P) in the OID: .1.3.6.1.2.1.P.22.1.1.X.A.B.C.D
note: we still dont know P in the above oid, infact we dont know if we 
have even received that or not...*/
			next_oid[6] = 4;
			next_oid[7] = 22;
			next_oid[8] = 1;
			next_oid[10] = 1; //ifIndex;
			next_oid[11] = i1;
			next_oid[12] = i2;
			next_oid[13] = i3;
			next_oid[14] = i4;


			/*printf("[GetNext..] nextoid: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
			  next_oid[0], next_oid[1], next_oid[2], next_oid[3],
			  next_oid[4], next_oid[5], next_oid[6], next_oid[7],
			  next_oid[8], next_oid[9], next_oid[10], next_oid[11],
			  next_oid[12], next_oid[13], next_oid[14], next_oid[15]); */
			/*all's left is next_oid[9]...if the following checks out, next_oid is good!*/
			switch (*(next_oid+9)) {

				case 1: //ipNetToMediaIfIndex
					//printf("[GetNext..]ipfn: ifIndex matched!\n");
					sprintf(data, "%d", 1);
					fclose (in);
					break;

				case 2: //ipNetToMediaPhysAddress
					sprintf(data, "%x:%x:%x:%x:%x:%x", 
							mac_addr[0], mac_addr[1], mac_addr[2], 
							mac_addr[3], mac_addr[4], mac_addr[5]);
					fclose (in);
					break;

				case 3: //ipNetToMediaNetAddress
					sprintf(data, "%d.%d.%d.%d", 
							next_oid[11], next_oid[12], 
							next_oid[13], next_oid[14]);
					fclose (in);
					break;

				case 4: //ipNetToMediaType
					sprintf(data, "%d", 1);
					fclose (in);
					break;

				default:
					/* this means the client did not send this info
					   thus, we'll respond with the first oid in the table.
					   that is ipNetToMediaIfIndex*/
					next_oid[9] = 1;
					sprintf(data, "%d", 1);
					fclose (in);
					/*printf("[IpFn] nextoid: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
					  next_oid[0], next_oid[1], next_oid[2], next_oid[3],
					  next_oid[4], next_oid[5], next_oid[6], next_oid[7],
					  next_oid[8], next_oid[9], next_oid[10], next_oid[11],
					  next_oid[12], next_oid[13], next_oid[14], next_oid[15]);*/
					break;
			}

		}
		else { /*interesting else*/
			/* THIS IS THE MOST INTERESTING PART OF THE CODE 
			 * WE ARE DEALING WITH A GETNEXT FOR A COMPLETE OID
			 * THIS MEANS IT CAN BE FROM ANYWHERE WITHIN THE OID TABLE
			 *
			 *proc/net/arp looks like:
			 *r2d3_knail1-> more /proc/net/arp
			 IP address       HW type     Flags       HW address            Mask     Device
			 192.168.1.1      0x1         0x2         00:04:5A:D2:90:45     *        wlan0
			 192.168.1.25     0x1         0x2         00:01:02:91:86:78     *        wlan0

			 * the output would look like:
			 * .ipNetToMediaIfIndex.1.10.5.1.1 = 3
			 * .ipNetToMediaIfIndex.1.10.5.1.15 = 3
			 * .ipNetToMediaPhysAddress.1.10.5.1.1 =  Hex: 00 E0 1E BC 04 8A 
			 * .ipNetToMediaPhysAddress.1.10.5.1.15 =  Hex: 00 D0 58 F6 0D 01 
			 * .ipNetToMediaNetAddress.1.10.5.1.1 = IpAddress: 10.5.1.1
			 * .ipNetToMediaNetAddress.1.10.5.1.15 = IpAddress: 10.5.1.15
			 * .ipNetToMediaType.1.10.5.1.1 = dynamic(3)
			 * .ipNetToMediaType.1.10.5.1.15 = other(1)
			 */

			/*printf("[FullOid Matched..] reqoid rcvd: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
			  req_oid[0], req_oid[1], req_oid[2], req_oid[3],
			  req_oid[4], req_oid[5], req_oid[6], req_oid[7],
			  req_oid[8], req_oid[9], req_oid[10], req_oid[11],
			  req_oid[12], req_oid[13], req_oid[14], req_oid[15]); */

			ReadFileIntoLL(ipNetToMediaTable_LL, req_oid, oid_length, next_oid, data);	

			/*NextOidCreatorAndDataPopulator(ipNetToMediaTable_LL,
			  req_oid, oid_length, next_oid, data);*/

		} /*end of interesting_else */

	} /*end of else if (req_type = snmpgetnext */
	else if (req_type == SNMPSET) //snmpset
	{
		int set_val_i;
		/*the req_oid better comply to two things here:
		 * 1. the oid must be c_ipNetToMediaType, and
		 * 2. and the oid must be complete, and have a valid ip address
		 * in it 
		 * 3. the set_val must be 2*/
		if (req_oid[9] != 4)
		{
			sprintf(data, "%d", -1); /* request made on bad oid*/
			return;
		}
		set_val_i = atoi(set_val);
		//	printf("setval s: %sd: %d", set_val, set_val_i);
		if (set_val_i != 2)
		{
			sprintf(data, "%d", -3); /* request made with bad set val*/
			return;
		}

		sprintf(ip_addr_c, "%d.%d.%d.%d", req_oid[11], req_oid[12], 
				req_oid[13], req_oid[14]);


		memset((char *) &req, 0, sizeof(req));

		if ((serv_hostent = gethostbyname(ip_addr_c)) == NULL) { 
			perror("gethostbyname error");
			return;                                
		}

		if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			perror("[IpFn] could not bind socket for deleting arp entry ");
			return;
		}

		sin = (struct sockaddr_in *) &req.arp_pa;

		bzero(sin,sizeof(struct sockaddr_in));
		sin->sin_family = AF_INET;

		memcpy(&sin->sin_addr, serv_hostent->h_addr, serv_hostent->h_length);

		if ((ioctl(sockfd, SIOCDARP, &req) < 0)) {
			perror("could not delete arp entry");
			return;
		}

		sprintf(data, "%d", 2);  /* 2 means invalidated as defined in RFC1213*/
		NullNextOid(next_oid);
		OidCopier(next_oid, req_oid, oid_length);/*return the same oid in next_oid*/
	} /*end of snmpset */

} /* end of ipfn */

int NextOidCreatorAndDataPopulator(ip_table *global_table, 
		const oid *req_oid, const int oid_length, 
		oid *next_oid, char *data)
{
	/* the two things of importance to this control block are
	 * (a) req_oid[9] to find out which of the 4 subtrees above and
	 * (b) ip address req_oid[11] thru req_oid[14].
	 * using the above, it can locate the current position in the tree 
	 * and parse the populated ipNetToMediaTable_LL to retrieve the relevant
	 * data.
	 */
	char ip_addr_string[15];
	char mac_addr_string[17], temp_mac[17];
	ip_table *anchor;
	int comp_val;

	bzero(data, sizeof(data));

	/*printf("[NextOidC..] reqoid rcvd: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
	  req_oid[0], req_oid[1], req_oid[2], req_oid[3],
	  req_oid[4], req_oid[5], req_oid[6], req_oid[7],
	  req_oid[8], req_oid[9], req_oid[10], req_oid[11],
	  req_oid[12], req_oid[13], req_oid[14], req_oid[15]); */
	/*first let's create the ip (in the getnext request) into a string*/
	sprintf(ip_addr_string, "%d.%d.%d.%d", 
			req_oid[11], req_oid[12], req_oid[13], req_oid[14]);

	/*let's also anchor the starting position of the linked list*/
	anchor = global_table;


	switch (req_oid[9])
	{
		case 1:
			/*either the next IfIndex or the first PhysAddress should be
			 * returned*/
			/*printf("ip addr st %s", ip_addr_string);
			  printf("globaltable->netaddress %s", global_table->NetAddress); */
			/*iterate until you find the right node in the ll*/
			while (((comp_val = strcmp(global_table->NetAddress, ip_addr_string)) != 0)
					&& (global_table->next_node->next_node != NULL))
			{
				/*printf("didnt match, iterate!, ip/mac in node %s/%s\n", 
				  global_table->NetAddress,
				  global_table->PhysAddress); */
				global_table = global_table->next_node;
			}

			if (global_table->next_node->next_node == NULL)	{
				//printf("globaltables next_node is NULL\n");
				/*this means we have reached the last line in proc/net/arp
				 * or in other words the last line inthe IfIndex subtree,
				 * so we have to send out the "first" linked lists 
				 * PhysAddress, and set oid accordingly*/

				global_table = anchor; /*the first ll*/
				strcpy(data, global_table->PhysAddress);
				/*printf("ip/mac : %s/%s, data: %s\n",
				  global_table->NetAddress,
				  global_table->PhysAddress, data); */


				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid, oid_length);
				next_oid[9] = 2;

				PutIpInNextOid(global_table->NetAddress, next_oid);
			} else if ((comp_val == 0) &&
					(global_table->next_node->next_node != NULL))
			{
				//printf("global_tables next node is not null, but we have a match\n");
				/*this means we matched the IP address in the OID with
				 * something within the ipNetToMediaTable and still haven't
				 * reached the end; thus we need to deliver the next value from
				 * within the SAME subtree (i.e. IfIndex) */

				global_table = global_table->next_node;
				/*strcpy(global_table->IfIndex, data);*/
				sprintf(data, "%d", global_table->IfIndex);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid,oid_length);
				/*note that next_oid[9] remains the same here
				  : we are still in the same subtree */

				PutIpInNextOid(global_table->NetAddress, next_oid);
			}
			break;
		case 2:
			/*either the next PhysAddress or the first NetAddress should
			 * be returned*/

			/*iterate until you find the right node in the ll*/
			while (((comp_val = strcmp(global_table->NetAddress, ip_addr_string)) != 0)
					&& (global_table->next_node->next_node != NULL))
				global_table = global_table->next_node;

			if (global_table->next_node->next_node == NULL)	{
				/*this means we have reached the last line in proc/net/arp
				 * or in other words the last line inthe PhysAddress subtree,
				 * so we have to send out the "first" linked lists 
				 * NetAddress, and set oid accordingly*/
				global_table = anchor; /*the first ll*/
				strcpy(data, global_table->NetAddress);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid, oid_length);
				next_oid[9] = 3;

				PutIpInNextOid(global_table->NetAddress, next_oid);
			} else if ((comp_val == 0) &&
					(global_table->next_node->next_node != NULL))
			{
				/*this means we matched the IP address in the OID with
				 * something within the ipNetToMediaTable and still haven't
				 * reached the end; thus we need to deliver the next value from
				 * within the SAME subtree (e.g. PhysAddress) */

				global_table = global_table->next_node;
				strcpy(data, global_table->PhysAddress);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid,oid_length);
				/*note that next_oid[9] remains the same here
				  : we are still in the same subtree */

				PutIpInNextOid(global_table->NetAddress, next_oid);
			}
			break;
		case 3: 
			/*either the next NetAddresss or the first Type should be
			 * returned*/
			/*iterate until you find the right node in the ll*/
			while (((comp_val = strcmp(global_table->NetAddress, ip_addr_string)) != 0)
					&& (global_table->next_node->next_node != NULL))
				global_table = global_table->next_node;

			if (global_table->next_node->next_node == NULL)	{
				/*this means we have reached the last line in proc/net/arp
				 * or in other words the last line inthe NetAddress subtree,
				 * so we have to send out the "first" linked lists 
				 * Type, and set oid accordingly*/
				global_table = anchor; /*the first ll*/
				/*strcpy(global_table->Type, data);*/
				sprintf(data, "%d", global_table->Type);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid, oid_length);
				next_oid[9] = 4;
				PutIpInNextOid(global_table->NetAddress, next_oid);

			} else if ((comp_val == 0) &&
					(global_table->next_node->next_node != NULL))
			{
				/*this means we matched the IP address in the OID with
				 * something within the ipNetToMediaTable and still haven't
				 * reached the end; thus we need to deliver the next value from
				 * within the SAME subtree (ie NetAddress) */

				global_table = global_table->next_node;
				strcpy(data, global_table->NetAddress);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid, oid_length);
				/*note that next_oid[9] remains the same here
				  : we are still in the same subtree */

				PutIpInNextOid(global_table->NetAddress, next_oid);
			}
			break;
		case 4:
			//printf("[IpFn]: how did you get here?\n");
			/*either the next Type or -2 (end of tree) should be returned
			*/
			/*iterate until you find the right node in the ll*/
			while (((comp_val = strcmp(global_table->NetAddress, ip_addr_string)) != 0)
					&& (global_table->next_node->next_node != NULL))
				global_table = global_table->next_node;

			if (global_table->next_node->next_node == NULL)	{
				/*this means we have reached the last line in proc/net/arp
				 * or in other words the last line inthe Type subtree,
				 that means end of the whole tree! just return -2 to agent_core
				 no need to mess around with next_oid;*/
				sprintf(data, "%d", -2);
			} else if ((comp_val == 0) &&
					(global_table->next_node->next_node != NULL))
			{
				/*this means we matched the IP address in the OID with
				 * something within the ipNetToMediaTable and still haven't
				 * reached the end; thus we need to deliver the next value from
				 * within the SAME subtree (ie Type) */

				global_table = global_table->next_node;
				/*strcpy(global_table->Type, data);*/
				sprintf(data, "%d", global_table->Type);

				NullNextOid(next_oid);
				OidCopier(next_oid, req_oid,oid_length);
				/*note that next_oid[9] remains the same here
				  : we are still in the same subtree */

				PutIpInNextOid(global_table->NetAddress, next_oid);
			}
			break;

		default: /*incorrect OID*/
			sprintf(data, "%d", -1);
			break;
	}
	1;
}

int PutIpInNextOid(char *ip_string, oid *next_oid) {
	/*this function changes the ip_string into four decimal numbers
	 * and sets next_oid[11] through [14] to them. */
	sscanf(ip_string, "%d.%d.%d.%d", &next_oid[11], &next_oid[12],
			&next_oid[13], &next_oid[14]);
	1;
}

int ReadFileIntoLL(ip_table *global_table, const oid *req_oid, int oid_length, 
		oid *next_oid, char *data)	{
	/*this would read /proc/net/arp into the passed global ipNetToMediaTable_LL 
	 * linked list*/
	char line[128];
	int i1, i2, i3, i4, m1, m2, m3, m4, m5, m6;
	char mac_addr_c[17];
	char ip_addr_c[15];
	char ifname[20];
	FILE *in = fopen ("/proc/net/arp", "r");
	global_table = (ip_table *)malloc(sizeof(ip_table));

	if (in == NULL)	{
		perror("IpFn:ReadFileIntoLL: ipNetToMediaTableLL malloc failed");
		return -1;
	}
	if (global_table == NULL) {
		perror("IpFn:ReadFileIntoLL: ipNetToMediaTableLL malloc failed");
		return -1;
	}

	global_table->prev_node = NULL;
	global_table->next_node = NULL;

	while (line == fgets(line, sizeof(line), in)) 
	{
		if (11 == sscanf (line, "%d.%d.%d.%d 0x%*x 0x%*x %x:%x:%x:%x:%x:%x %*[^ ] %20s\n",
					&i1, &i2, &i3, &i4,
					&m1, &m2, &m3, &m4, &m5, &m6, ifname))
		{
			sprintf(mac_addr_c, "%x:%x:%x:%x:%x:%x", m1,m2,m3,m4,m5,m6);
			sprintf(ip_addr_c, "%d.%d.%d.%d", i1,i2,i3,i4);
			//	printf("[ReadLine..]: mac addr is %s\n", mac_addr_c);
			push_ip(global_table, ip_addr_c, mac_addr_c);
		}
		else 
			continue;
	}


	/*printf("[ReadLine..] reqoid rcvd: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
	  req_oid[0], req_oid[1], req_oid[2], req_oid[3],
	  req_oid[4], req_oid[5], req_oid[6], req_oid[7],
	  req_oid[8], req_oid[9], req_oid[10], req_oid[11],
	  req_oid[12], req_oid[13], req_oid[14], req_oid[15]); */

	NextOidCreatorAndDataPopulator(global_table,
			req_oid, oid_length, next_oid, data);
	PrintAll(global_table);

	1;
}


int PrintAll(ip_table *tabul)	{
	ip_table *tab;
	//tab = (ip_table *)malloc(sizeof(ip_table));
	//memset(tab, 0, sizeof(ip_table));

	tab = tabul;
	while (tab->next_node != NULL)
	{
		//printf("[PrintAll] ip %s, mac, %s\n", tab->NetAddress, tab->PhysAddress);
		tab = tab->next_node;
	}
	//printf("[nullPrintAll] ip %s, mac, %s\n", tab->NetAddress, tab->PhysAddress);

	1;
}
int push_ip(ip_table *tabul, char *ip_addr, char *mac_addr)
{
	/*this fn pushes the newly recieved ip and mac address in the
	 * ip doubly linked list. note that it assumes that the ipNetToMediaTable_LL
	 * has been malloc'ed with its prev_node and next_node set to NULL*/
	ip_table *current, *new;
	new = (ip_table *)malloc(sizeof(ip_table));
	memset(new, 0, sizeof(ip_table));

	current = tabul;
	while(current->next_node != NULL)
		current = current->next_node;

	strcpy(current->NetAddress, ip_addr);
	strcpy(current->PhysAddress, mac_addr);

	//printf("[push_ip]: mac addr from table itself is %s\n", current->PhysAddress);
	current->IfIndex = 1;
	current->Type = 1;

	current->next_node = new;
	current->next_node->next_node = NULL;
	current->next_node->prev_node = current;
	1;
} /*end of push*/

