/* this file defines the various typedefs and constant enumerations used by various subsystem components of the engine */

#include <sys/types.h>

typedef u_long oid;
/* used as:
 * oid sysDescr[9] = {1 3 6 1 2 1 1 1 0};
 * as sysDescr.0 is : .1.3.6.1.2.1.1.1.0
 */

#define VERSION 0
#define SNMPGET 0
#define SNMPGETNEXT 1
#define SNMPGETRESPONSE 2
#define SNMPSET 3
#define GN_CACHE_TIME 5
#define ERR_NOERROR 0
#define ERR_TOOBIG	1
#define ERR_NOSUCHNAME	2
#define ERR_BADVALUE	3
#define ERR_READONLY	4
#define ERR_GENERR	5

#define PROC_TCP_ESTABLISHED 1
#define PROC_TCP_SYN_SENT 2
#define PROC_TCP_SYN_RECV 3
#define PROC_TCP_FIN_WAIT1 4
#define PROC_TCP_FIN_WAIT2 5
#define PROC_TCP_TIME_WAIT 6
#define PROC_TCP_CLOSE 7
#define PROC_TCP_CLOSE_WAIT	8
#define PROC_TCP_LAST_ACK	9
#define PROC_TCP_LISTEN	10
#define PROC_TCP_CLOSING 11

#define SNMP_TCP_ESTABLISHED 5
#define SNMP_TCP_SYN_SENT 3
#define SNMP_TCP_SYN_RECV 4
#define SNMP_TCP_FIN_WAIT1 6
#define SNMP_TCP_FIN_WAIT2 7
#define SNMP_TCP_TIME_WAIT 11
#define SNMP_TCP_CLOSE 1
#define SNMP_TCP_CLOSE_WAIT	8
#define SNMP_TCP_LAST_ACK	9
#define SNMP_TCP_LISTEN	2
#define SNMP_TCP_CLOSING 10
#define SNMP_TCP_DELETE_TCB 12
// definition of datatypes.
//typedef char	DisplayString [255]
//typedef int 	TimeTicks
//typedef int		Integer
//commented out as this agent doesnt really use datatypes.
//the definitions for each oid are defined in oids_supported.h
//which would only be useful while ber encoding.
typedef struct  {
	/*struct returned by query_sanity_checker to main() */
	short int error_status;
	// short int error_index;
	int request_id;
	// int *array_of_pointer_to_oids; /*where an oid is an array of integers*/
	oid requested_oid[20]; //array of u_long, filled with -1's for N/A branches
	int oid_length;  /* the length of valid oid */
	short int request_type;
	char snmpset_value[128];
} checked_query_result;

typedef struct 	{
	/* this is the struct returned by the ber_decoder. 
	 * as you can see, all the fields in this struct match the 
	 * incoming request packet's fields */
	short int version;
	char community[32];
	short int pdu_type;
	int request_id;
	short int error_status;
	// short int error_index;
	// int *array_of_pointer_to_oids; /*where an oid is a string */
	oid OID[20]; //anticipating an oid of type string from ber_decoder
	char snmpset_value[128]; // the value to set if this is an snmpset request 
} decoded_packet_struct;

typedef struct  {

	short int version;
	char community[32];
	short int pdu_type;
	int request_id;
	short int error_status;
	short int error_index;
	char Oid[100]; /*dont anticipate it to be more than 66 max*/
	char Value[128];

} packet_to_be_encoded;

typedef struct 	{
	oid requested_oid[20]; /*20 is max*/
	//char *datatype;	
	/*these types would be typedef'd in lib/snmpd_types.h */
	char value[128];
	/*note: this value is returned as a string.
	 * the ber_encoder would have to cast this value as 
	 * specified in the datatype */
} oid_datatype_value;

typedef struct ip_table {
	struct ip_table *next_node;
	struct ip_table *prev_node;
	int IfIndex;
	char PhysAddress[17];
	char NetAddress[15];
	int Type;
} ip_table;


typedef struct tcp_table {
	struct tcp_table *next_node;
	struct tcp_table *prev_node;
	int tcpConnState;
	char LocalAddress[15];
	int LocalPort;
	char RemoteAddress[15];
	int RemotePort;
} tcp_table;

void AgentCore(short int req_type, 
		const oid *req_oid, 
		int oid_length,
		char *set_value, 
		oid_datatype_value *oid_and_data); 


void SystemFn (short int req_type,const oid *req_oid, 
		const int oid_length, char *data, oid *next_oid,  char *set_val);
void IcmpFn (short int req_type,const oid *req_oid, 
		const int oid_length, char *data, oid *next_oid);
void IpFn (short int req_type, const oid *req_oid, const int oid_length, 
		char *data, oid *next_oid, char *set_val);
int IpOidMatch (char *line, const oid *req_oid, int *mac_addr);

void TcpFn (short int req_type,const  oid *req_oid, const int oid_length,
		char *data, oid *next_oid);

int QuerySanityChecker (const decoded_packet_struct *decoded_request, 
		checked_query_result *qsn_output);

int CreateErrorMessage (const decoded_packet_struct *decoded_request,
		checked_query_result *qsn_output,
		packet_to_be_encoded *response);

int CreateGoodMessage (const decoded_packet_struct *decoded_request,
		const oid_datatype_value *oid_and_data,
		packet_to_be_encoded *response);



int RequestDecoder(const char *req, decoded_packet_struct *dps);
int RequestEncoder(const packet_to_be_encoded *p2be, char *output_string);
int RequestEncoder(const packet_to_be_encoded *p2be, char *output_string);

int main();

