/* query sanity checker */
/* from RFC 1157:
 *
 *
 *INTEGER
 *ErrorStatus ::=
 *INTEGER {
 *noError(0),
 *tooBig(1),
 *noSuchName(2),
 *badValue(3),
 *readOnly(4)
 *genErr(5)
 *endofmib(6) <-custom
 *}
 */

#include <stdio.h>
#include <string.h>
#include "lib/oids_supported.h" /*takes care of including snmpd_types.h*/


int QuerySanityChecker	(const decoded_packet_struct *decoded_request,
		checked_query_result *qsn_output)
{
	int ret_val = 0;

	if ((ret_val = VersionChecker(decoded_request, qsn_output)) < 0)
		return(ret_val);

	if ((ret_val = Community_and_PDU_type_Checker(decoded_request, qsn_output)) < 0)
		return(ret_val);

	if ((ret_val = OidChecker(decoded_request, qsn_output)) < 0)
		return(ret_val);

	/*for all the requests which are pristine pure, and valid*/
	CreateGoodCheckQueryResult(decoded_request, qsn_output);
	return(ret_val);
}

int CreateGoodCheckQueryResult(const decoded_packet_struct *dr, checked_query_result *cqr)
{
	char temp_charOID[200];
	bzero(temp_charOID, sizeof(temp_charOID));


	cqr->error_status = 0;
	cqr->request_id = dr->request_id;
	/*requested_oid*/
	QNullNextOid(cqr->requested_oid);
	memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
	/*requested oid length; note: i'm anticipating that BerDecoder filled 
	 * -1's in all the N/A oid fields for the requested_oid*/
	bzero(temp_charOID, sizeof(temp_charOID));
	cqr->oid_length = QInt2CharOID(dr->OID,temp_charOID); 
	cqr->request_type = dr->pdu_type;
	strcpy(cqr->snmpset_value, dr->snmpset_value);
	1;
}

int OidChecker(const decoded_packet_struct *dr, checked_query_result *cqr)
{
	/*responsibilities (items):
	 * 1. if SNMPGET/SET : check if oid length is complete depending on
	 * req_oid[9]
	 * 2. if GETNEXT, depending on oid length, create the next oid, and put
	 * that in cqr..
	 * 3. ensuring that the first 6 branches in the OID are always 
	 * .1.3.6.1.2.1
	 * 4. (get/set) ensuring for ip and tcp, that the fixed oid after req_oid[9] is always:
	 * 22.1 (ip) and
	 * 13.1 (tcp) respectively
	 * (you see, the rest of the OID either gets verified up top, or in the 
	 * respective fn's in agent_core 
	 *
	 * 5. (get/set) ensure  .0is  at the end of  icmpfn and systemfn*
	 * 6. (set) if the oid attemped to be set is settable or not *
	 *
	 * NOTE: I'll let OID lengths that match for get/set but are incorrect
	 * after the first six branches (e.g. 1.3.6.1.2.1.5."6".0)..i'll let 
	 * them slip through, because agent core is going to catch all such s
	 * stuff */

	char temp_charOID[200];
	int oid_l;


	/* now to check if snmpset is being done on settable oid 
	 * we only support: 
	 * c_sysContact0
	 * c_sysLocation0
	 * c_ipNetToMediaType */

	if (dr->pdu_type == SNMPSET)
	{
		switch (dr->OID[6]) 
		{
			case 4:
				if  (dr->OID[9] != 4)
				{
					cqr->error_status = ERR_READONLY;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;
			case 1:
				if  (dr->OID[7] == 4) 
					break;
				else if (dr->OID[7] == 6)
					break;
				else
				{
					printf("oid7 %d, get here\n", dr->OID[7]);
					cqr->error_status = ERR_READONLY;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;

			default:
				cqr->error_status = ERR_READONLY;
				cqr->request_id = dr->request_id;
				memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
				cqr->oid_length = 0; /*really doesnt matter*/
				cqr->request_type = dr->pdu_type;
				strcpy(cqr->snmpset_value, dr->snmpset_value);
				return(-1);
				break;	
		}
	}



	/*-----now, checking for snmpget/snmpset queries, and
	 * respective lengths,and  for ip(22.1) and tcp(13.1) and
	 * also for the last system(.0) and icmp(.0)---*/
	if ((dr->pdu_type == SNMPGET) || (dr->pdu_type == SNMPSET))
	{
		/*-----checking for 1.3.6.1.2.1 ------- */
		if (!((dr->OID[0] == 1) && (dr->OID[1] == 3) && (dr->OID[2] == 6)
					&& (dr->OID[3] == 1) && (dr->OID[4] == 2) && (dr->OID[5] == 1)))
		{
			cqr->error_status = ERR_NOSUCHNAME;
			cqr->request_id = dr->request_id;
			memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
			cqr->oid_length = 0; /*really doesnt matter*/
			cqr->request_type = dr->pdu_type;
			strcpy(cqr->snmpset_value, dr->snmpset_value);
			return(-1);
		}
		switch(dr->OID[6])
		{
			case 1:
				bzero(temp_charOID, sizeof(temp_charOID));
				if (QInt2CharOID(dr->OID, temp_charOID) != 9)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				if (dr->OID[8] != 0)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;

			case 4:
				bzero(temp_charOID, sizeof(temp_charOID));
				if (QInt2CharOID(dr->OID, temp_charOID) != 15)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				if (!((dr->OID[7] == 22) && (dr->OID[8] == 1)))
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;

			case 5:
				bzero(temp_charOID, sizeof(temp_charOID));
				if (QInt2CharOID(dr->OID, temp_charOID) != 9)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				if (dr->OID[8] != 0)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;

			case 6:
				bzero(temp_charOID, sizeof(temp_charOID));
				if (QInt2CharOID(dr->OID, temp_charOID) != 20)
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				if (!((dr->OID[7] == 13) && (dr->OID[8] == 1)))
				{
					cqr->error_status = ERR_NOSUCHNAME;
					cqr->request_id = dr->request_id;
					memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
					cqr->oid_length = 0; /*really doesnt matter*/
					cqr->request_type = dr->pdu_type;
					strcpy(cqr->snmpset_value, dr->snmpset_value);
					return(-1);
				}
				break;
		} /*end of switch */
	} /*end of if snmpget/snmpset */


	/*-----finally checking the snmpget request:
	 * note: this might not result in a bad query----*/

	if (dr->pdu_type == SNMPGETNEXT)
	{
		bzero(temp_charOID, sizeof(temp_charOID));
		oid_l = QInt2CharOID(dr->OID, temp_charOID);
		//printf("oid_l %d\n", oid_l);

		/*-----checking for 1.3.6.1.2.1 ------- */
		if ((!((dr->OID[0] == 1) && (dr->OID[1] == 3) && (dr->OID[2] == 6)
						&& (dr->OID[3] == 1) && (dr->OID[4] == 2) && (dr->OID[5] == 1))) 
				&& (oid_l < 6))
		{
			/*the OID received is too small, return sysDescr0*/
			cqr->error_status = ERR_NOERROR;
			cqr->request_id = dr->request_id;
			QNullNextOid(cqr->requested_oid);
			memcpy(cqr->requested_oid, c_sysDescr0, sizeof(c_sysDescr0));
			cqr->oid_length = sizeof(c_sysDescr0)/4;
			cqr->request_type = 0;
			strcpy(cqr->snmpset_value, dr->snmpset_value);
			return(-1);
		}

		switch (oid_l)
		{
			case 1: case 2: case 3: case 4: case 5: case 6:
				/*OId too small, so return, sysDescr0 */
				cqr->error_status = ERR_NOERROR;
				cqr->request_id = dr->request_id;
				QNullNextOid(cqr->requested_oid);
				memcpy(cqr->requested_oid, dr->OID, 4*oid_l);
				cqr->oid_length = oid_l; /* u_long = 4B*/
				cqr->request_type = dr->pdu_type;
				strcpy(cqr->snmpset_value, dr->snmpset_value);
				DpsPrinter("qsn", dr);
				CqrPrinter("qsn", cqr);
				return(-1); /*not really an error, but returning a
							  -1 so that QuerySanityChecker aborts */
				break;

			default: /*anything greater and length of 6 is specific to
					   tree (e.g. system, ip .. etc) */

				/*first, let's check lengths again, and if the lengths match
				 * dont mess around with the OID.., if they are NOT,
				 * then return the first OId in that tree */

				switch(dr->OID[6])
				{
					case 1:
						bzero(temp_charOID, sizeof(temp_charOID));
						if (QInt2CharOID(dr->OID, temp_charOID) != 9)
						{
							cqr->error_status = ERR_NOERROR;
							cqr->request_id = dr->request_id;
							QNullNextOid(cqr->requested_oid);
							memcpy(cqr->requested_oid, c_sysDescr0, sizeof(c_sysDescr0));
							cqr->oid_length = sizeof(c_sysDescr0)/4;
							cqr->request_type = 0;
							strcpy(cqr->snmpset_value, dr->snmpset_value);
							return(-1);
						}
						break;

					case 4:
						bzero(temp_charOID, sizeof(temp_charOID));
						if (QInt2CharOID(dr->OID, temp_charOID) != 15)
						{
							cqr->error_status = ERR_NOERROR;
							cqr->request_id = dr->request_id;
							QNullNextOid(cqr->requested_oid);
							memcpy(cqr->requested_oid, c_ipNetToMediaIfIndex, 
									sizeof(c_ipNetToMediaIfIndex));
							cqr->oid_length = sizeof(c_ipNetToMediaIfIndex)/4;
							cqr->request_type = 1;
							strcpy(cqr->snmpset_value, dr->snmpset_value);
							return(-1);
						}
						break;

					case 5:
						bzero(temp_charOID, sizeof(temp_charOID));
						if (QInt2CharOID(dr->OID, temp_charOID) != 9)
						{
							cqr->error_status = ERR_NOERROR;
							cqr->request_id = dr->request_id;
							QNullNextOid(cqr->requested_oid);
							//memcpy(cqr->requested_oid, dr->OID, 4*len);
							memcpy(cqr->requested_oid, c_icmpInEchos0, 
									sizeof(c_icmpInEchos0));
							cqr->oid_length = sizeof(c_icmpInEchos0);
							cqr->request_type = 0;
							/* by making a snmpget request we would be assured
							 * that the first oid and data value is returned*/
							strcpy(cqr->snmpset_value, dr->snmpset_value);
							return(-1);
						}
						break;

					case 6:
						bzero(temp_charOID, sizeof(temp_charOID));
						if (QInt2CharOID(dr->OID, temp_charOID) != 20)
						{
							cqr->error_status = ERR_NOERROR;
							cqr->request_id = dr->request_id;
							QNullNextOid(cqr->requested_oid);
							memcpy(cqr->requested_oid, c_tcpConnState, 
									sizeof(c_tcpConnState));
							cqr->oid_length = sizeof(c_tcpConnState)/4;
							cqr->request_type = 1;
							strcpy(cqr->snmpset_value, dr->snmpset_value);
							return(-1);
						}
						break;
					default:
						/*oid not supported*/
						cqr->error_status = ERR_NOSUCHNAME;
						cqr->request_id = dr->request_id;
						QNullNextOid(cqr->requested_oid);
						memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
						cqr->oid_length = 0;
						cqr->request_type = dr->pdu_type;
						strcpy(cqr->snmpset_value, dr->snmpset_value);
						return(-1);
						break;
				} /*end of switch QInt2CharOID */
				break; /*for default */
		} /*end of switch oid_l */
	} /*end of snmpgetnext check*/

	return(0);
} /*end of OidChecker */

int Community_and_PDU_type_Checker(const decoded_packet_struct *dr, 
		checked_query_result *cqr)
{
	FILE *snmpd_cfg;
	char line2[40];
	char *conta;

	switch(dr->pdu_type)
	{
		case SNMPGET: case SNMPSET: case SNMPGETNEXT:
			break;

		default:
			cqr->error_status = ERR_GENERR;
			cqr->request_id = dr->request_id;
			QNullNextOid(cqr->requested_oid);
			memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
			cqr->oid_length = 0; /*really doesnt matter*/
			cqr->request_type = dr->pdu_type;
			strcpy(cqr->snmpset_value, dr->snmpset_value);
			return(-1);
			break;

	}

	if (!(snmpd_cfg = fopen("etc/snmpd.cfg", "r")))
	{
		perror("[Community_and_PDU_type_Checker]: cannot open etc/snmpd.cfg");
		return;
	}

	if (dr->pdu_type == SNMPSET)
	{
		while (line2 == fgets (line2, sizeof(line2), snmpd_cfg)) {
			conta = strtok (line2, ":");
			if (strcmp(conta,"RW_Community") == 0)
			{
				conta = strtok(NULL, "\0");
				break;
			}
		}
	} else if ((dr->pdu_type == SNMPGET) || (dr->pdu_type == SNMPGETNEXT))
	{
		while (line2 == fgets (line2, sizeof(line2), snmpd_cfg)) 
		{
			conta = strtok (line2, ":");
			if (strcmp(conta,"RO_Community") == 0)
			{
				conta = strtok(NULL, "\0");
				break;
			}
		}

		if (conta[strlen(conta)-1] == '\n')
			conta[strlen(conta)-1] = '\0';

		/*the RO commstring didnt match, let's see if rw commstring matches*/
		if (strcmp(dr->community, conta) != 0)
		{
			while (line2 == fgets (line2, sizeof(line2), snmpd_cfg)) {
				conta = strtok (line2, ":");
				if (strcmp(conta,"RW_Community") == 0)
				{
					conta = strtok(NULL, "\0");
					break;
				}
			}

		} else 
		{
			fclose(snmpd_cfg);
			return(0);
		}

	} /*end of if SNMPGET or SNMPGETNEXT*/

	if (conta[strlen(conta)-1] == '\n')
		conta[strlen(conta)-1] = '\0';


	if (strcmp(dr->community, conta) != 0)
	{
		cqr->error_status = ERR_NOSUCHNAME;
		cqr->request_id = dr->request_id;
		QNullNextOid(cqr->requested_oid);
		memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
		cqr->oid_length = 0; /*really doesnt matter*/
		cqr->request_type = dr->pdu_type;
		strcpy(cqr->snmpset_value, dr->snmpset_value);
		fclose(snmpd_cfg);
		return(-1);
	}


	fclose(snmpd_cfg);
	return(0);
}

int VersionChecker(const decoded_packet_struct *dr, checked_query_result *cqr)
{

	if (dr->version != 0)
	{
		cqr->error_status = ERR_NOSUCHNAME;
		cqr->request_id = dr->request_id;
		QNullNextOid(cqr->requested_oid);
		memcpy(cqr->requested_oid, dr->OID, sizeof(dr->OID));
		cqr->oid_length = 0; /*really doesnt matter*/
		cqr->request_type = dr->pdu_type;
		strcpy(cqr->snmpset_value, dr->snmpset_value);

		return(-1);
	}
	else
		return(0);
}


int QNullNextOid(oid *next_oid)	{
	/*this function initializes all values in the next_oid array*/
	int i;
	for (i=6;i<20;i++)	
		next_oid[i] = -1;
	/* making all -1 */
	1;
}

int QInt2CharOID(const oid *int_oid,char *char_OID)
{
	/* note: the non valid oid branches would be marked as -1*/
	int i, oid_length;
	char c_tmp[10];
	for (i=0; i<20; i++)
	{
		if (int_oid[i] != -1)
		{
			sprintf(c_tmp, "%d.", int_oid[i]);
			strcat(char_OID, c_tmp);
		} else
		{
			if (char_OID[strlen(char_OID)-1] == '.')
				char_OID[strlen(char_OID)-1] = '\0';
			//printf("final string:%sEND\n", char_OID);
			return(i);
		}
	}
	if (char_OID[strlen(char_OID)-1] == '.')
		char_OID[strlen(char_OID)-1] = '\0';
	//printf("final stringtcp:%sEND\n", char_OID);
	return(i);
}
/*
   int main()	{ 
   decoded_packet_struct dps;
   checked_query_result cqr;
   oid testoid[] = {1,3,6,1,2,1,5, 6,0};
//oid testoid[] = {1,3,6,1,2,1,4,22,1, 4,1,192,168,1,1};

dps.version = 0;
strcpy(dps.community, "public");
dps.pdu_type = SNMPGET;
dps.request_id = 1234;
dps.error_status = 0;
QNullNextOid(dps.OID);
memcpy(dps.OID, testoid, sizeof(testoid));
//memcpy(dps.OID, c_sysDescr0, sizeof(c_sysDescr0));
strcpy(dps.snmpset_value,"0");
printf("[MAIN.DPS] error_status: %d request_id: %d request_type: %d snmpset_value: %s oid_length: %d community: %sEND\n",
dps.error_status, dps.request_id, dps.pdu_type, 
dps.snmpset_value, sizeof(dps.OID), dps.community);

printf("[MAIN.DPS] requested_oid: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d %d.%d.%d.%d.%d %d.%d.%d.%d.%d\n", 
dps.OID[0], dps.OID[1], 
dps.OID[2], dps.OID[3],
dps.OID[4], dps.OID[5], 
dps.OID[6], dps.OID[7],
dps.OID[8], dps.OID[9],
dps.OID[10], dps.OID[11], 
dps.OID[12], dps.OID[13],
dps.OID[14], dps.OID[15],
dps.OID[16], dps.OID[17],
dps.OID[18], dps.OID[19]);


QuerySanityChecker(&dps, &cqr);


printf("[MAIN.QSC] error_status: %d request_id: %d request_type: %d snmpset_value: %s oid_length: %d\n",
cqr.error_status, cqr.request_id, cqr.request_type, 
cqr.snmpset_value, cqr.oid_length);
printf("[MAIN.QSC] requested_oid: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d %d.%d.%d.%d.%d %d.%d.%d.%d.%d\n", 
cqr.requested_oid[0], cqr.requested_oid[1], 
cqr.requested_oid[2], cqr.requested_oid[3],
cqr.requested_oid[4], cqr.requested_oid[5], 
cqr.requested_oid[6], cqr.requested_oid[7],
cqr.requested_oid[8], cqr.requested_oid[9],
cqr.requested_oid[10], cqr.requested_oid[11], 
cqr.requested_oid[12], cqr.requested_oid[13],
cqr.requested_oid[14], cqr.requested_oid[15],
cqr.requested_oid[16], cqr.requested_oid[17],
cqr.requested_oid[18], cqr.requested_oid[19]);

1;
}
*/
