/*main() for snmpd*/
#include <stdio.h>
#include <stdlib.h>
#include "lib/oids_supported.h"
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <unistd.h>
#include <getopt.h>


int main(int argc, char **argv) {
	decoded_packet_struct dps;
	checked_query_result cqr;
	packet_to_be_encoded response;
	oid_datatype_value oid_and_data;
	char request[200];	
	struct sockaddr_in fsin;
	unsigned int alen;
	int bytes_rcvd, socko;
	char output_string[250];
	oid tmp_oid[20];
	int port_number, i;
	static int verbose_flag = 0;
	static int port_specified = 0 ;
	int lop = 0;

	while (1)	{

		static struct option longopts[] =
		{
			{"verbose",0, &verbose_flag, 'v'},
			{"port",1, 0, 'p'},
			{"help",0,0,'h'},
			{0, 0, 0, 0}
		};

		i = getopt_long(argc, argv, "?hvp:", longopts,  &lop);


		if (i == -1)
			break;

		switch (i) {
			case 'v':
				verbose_flag++;
				break;
			case 'p':
				port_number = atoi(optarg);
				port_specified++;
				break;	
			case 'h':
			case '?':
			default:
				printf("[gmov-snmp v0.1] usage: \n snmpd [-h] [-p <portnumber>] [-v]\n");
				printf(" -p which UDP port to listen on (default 161)\n");
				printf(" -v to turn verbosity on\n");
				printf(" -h to print (this) help msg\n");
				return;
				break;
		}
	}

	if (!(port_specified)) 
		port_number = 161;

	printf("[gmov-snmpd v0.1] SNMP Daemon Started\n");
	sleep(1);
	printf("Attempting to listen on port %d..", port_number);


	alen = sizeof(fsin);
	if ((socko  = sock_bind(port_number)) < 1)
	{
		perror("\ncouldn't sock/bind");
		return;
	} else 
		printf("Ready\n");

	if (verbose_flag)
		printf("Verbose Mode on..\n");

	/*---endless loop starts here----*/
	while (1)
	{
		bzero(request, sizeof(request));
		if ((bytes_rcvd = recvfrom(socko, request, 
						sizeof(request), 0, (struct sockaddr *)&fsin, &alen)) < 0)
		{
			perror("recvfrom error");
		}


		memset(&dps, 0, sizeof(dps));
		RequestDecoder(request, &dps);
		/*Ber_Decoder(buf,&dps); */
		if (verbose_flag)
			DpsPrinter("Main", &dps);

		/*sending dps to qsn..*/
		memset(&cqr, 0, sizeof(cqr));
		QuerySanityChecker (&dps, &cqr);

		if (verbose_flag)
			CqrPrinter("Main", &cqr);

		if (cqr.error_status != 0) {
			/*something bad with the request */
			memset(&response, 0, sizeof(response));
			CreateErrorMessage(&dps, &cqr, &response);
			if (verbose_flag)
				printf("ouchie?\n");
			P2bEPrinter("Main", &response);
		} 
		else {
			/*request seems good, let's get the data*/
			memset(&oid_and_data, 0, sizeof(oid_and_data));

			AgentCore(	cqr.request_type, cqr.requested_oid, cqr.oid_length*4,
					cqr.snmpset_value, &oid_and_data);

			if (verbose_flag)
				OidAndDataPrinter("Main", &oid_and_data);
			if (atoi(oid_and_data.value) == -1)	{
				/*invalid request made*/
				cqr.error_status = ERR_NOSUCHNAME;
				memset(&response, 0, sizeof(response));
				CreateErrorMessage(&dps, &cqr, &response);
			} 
			else if (atoi(oid_and_data.value) == -2) {
				/*end of supported tree */
				cqr.error_status = 6;
				//CqrPrinter("MainAfterNull", &cqr);
				memset(&response, 0, sizeof(response));
				CreateErrorMessage(&dps, &cqr, &response);
			}
			else if (atoi(oid_and_data.value) == -3) {
				/*set attempted with incorrect value*/
				cqr.error_status = ERR_BADVALUE;
				CqrPrinter("MainAfterNull", &cqr);
				memset(&response, 0, sizeof(response));
				CreateErrorMessage(&dps, &cqr, &response);
			}
			else {
				/*everything checks out*/
				memset(&response, 0, sizeof(response));
				CreateGoodMessage(&dps, &oid_and_data, &response);
			}	
		}

		RequestEncoder(&response, output_string);

		if (verbose_flag)
			P2bEPrinter("Main", &response);

		/*if (verbose_flag)*/
		/*to write out the response in the non verbose format*/
		if (verbose_flag)
			printf("[Main] ResponseString:  %s\n\n\n", output_string);
		else 
			printf("[Main] ResponseString:  %s\n", output_string);

		/*BerEncoder(&response, return_buf); */

		if (sendto(socko, output_string, 
					strlen(output_string), 0, 
					(struct sockaddr *)&fsin, sizeof(fsin)) < 0)
			perror("couldn't send data");

	} /*end of while(1) loop */
	return;
} /*end of main() */




int sock_bind(int poart)	{
	struct sockaddr_in serv_addr;
	int lfd, diditbind, diditlisten, errno;

	lfd = socket (PF_INET, SOCK_DGRAM, 0);
	if (lfd < 0)
	{
		printf("\ncant create sock: %d\n", errno);
		return;
	}
	bzero(&serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons(poart);

	diditbind =  
		bind(lfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr));	
	if (diditbind < 0 )
	{
		perror("bind error");
		return -1;
	}

	return lfd;
}
