/* file handling SYSTEM branch of snmp tree */
#include <string.h>
#include <stdio.h>
#include "lib/oids_supported.h"


int SysDescrFn(char *data) 
{
	FILE *version, *hostname;
	char line1[128], line2[128];
	char os[15], vers[15], naam[32];


	if (!(version = fopen("/proc/version", "r")))
	{
		perror("[SysDescrFn]: cannot open /proc/version..");
		return;
	}
	if (!(line1 == fgets (line1, sizeof(line1), version))) 
	{
		perror("[SysDescrFn]: cannot read /proc/version..");
		return;
	}
	if (sscanf(line1, "%s %*s %s %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ] %*[^ ]", os, vers) != 2)
	{
		perror("[SysDescrFn]: cannot scan /proc/version..");
		return;
	}


	if (!(hostname = fopen("/proc/sys/kernel/hostname", "r")))
	{
		perror("[SysdescrFn]: cannot open /proc/sys/kernel/hostname");
		return;
	}

	if (!(line2 == fgets (line2, sizeof(line2), hostname)))
	{
		perror("[SysDescrFn]: cannot read /proc/sys/kernel/hostname..");
		return;
	}


	if (sscanf(line2, "%s", naam) != 1)
	{
		perror("[SysDescrFn]: cannot scan /proc/sys/kernel/hostname..");
		return;
	}

	sprintf(data, "%s: %s version %s", naam, os, vers);
	fclose(version);
	fclose(hostname);
	1;
}

int SysUptimeFn(char *data) 
{
	FILE *uptime;
	char line2[40];
	char Waqt[30];
	int waqt;

	if (!(uptime = fopen("/proc/uptime", "r")))
	{
		perror("[SysUpTimeFn]: cannot open /proc/uptime");
		return;
	}

	if (!(line2 == fgets (line2, sizeof(line2), uptime)))
	{
		perror("[SysUpTimeFn]: cannot read /proc/uptime");
		return;
	}

	if (sscanf(line2, "%d.%*2d %*d.%*2d", &waqt) != 1)
	{
		perror("[SysUpTimeFn]: cannot scan /proc/uptime");
		return;
	}
	sprintf(data, "%d", waqt);
	fclose(uptime);
	1;
}


int SysContactFnRO(char *data) 
{
	FILE *snmpd_cfg;
	char line2[40];
	char *conta;

	if (!(snmpd_cfg = fopen("etc/syscontact.cfg", "r")))
	{
		perror("[SysContactFnRO]: cannot open etc/syscontact.cfg");
		return;
	}

	while (line2 == fgets (line2, sizeof(line2), snmpd_cfg)) {
		conta = strtok (line2, ":");
		if (strcmp(conta,"sysContact") == 0)
		{
			conta = strtok(NULL, "\0");
			break;
		}

	}
	if (conta[strlen(conta)-1] == '\n')
		conta[strlen(conta)-1] = '\0';
	strcpy(data, conta);
	fclose(snmpd_cfg);

	1;
}

int SysNameFn(char *data) 
{
	FILE *hostname;
	char line2[128];
	char naam[32];
	if (!(hostname = fopen("/proc/sys/kernel/hostname", "r")))
	{
		perror("[SysNameFn]: cannot open /proc/sys/kernel/hostname");
		return;
	}

	if (!(line2 == fgets (line2, sizeof(line2), hostname)))
	{
		perror("[SysNameFn]: cannot read /proc/sys/kernel/hostname..");
		return;
	}


	if (sscanf(line2, "%s", naam) != 1)
	{
		perror("[SysNameFn]: cannot scan /proc/sys/kernel/hostname..");
		return;
	}

	sprintf(data, "%s", naam);
	fclose(hostname);
	1;
}


int SysLocationFnRO(char *data) 
{
	FILE *snmpd_cfg;
	char line2[40];
	char *conta;

	if (!(snmpd_cfg = fopen("etc/syslocation.cfg", "r")))
	{
		perror("[SysLocationFnRO]: cannot open etc/syslocation.cfg");
		return;
	}

	while (line2 == fgets (line2, sizeof(line2), snmpd_cfg)) {
		conta = strtok (line2, ":");
		if (strcmp(conta,"sysLocation") == 0)
		{
			conta = strtok(NULL, "\0");
			break;
		}

	}
	if (conta[strlen(conta)-1] == '\n')
		conta[strlen(conta)-1] = '\0';
	strcpy(data, conta);
	fclose(snmpd_cfg);

	1;
}

void SystemFn (short int req_type,const oid *req_oid,
		const int oid_length, char *data, oid *next_oid, char *set_val)
{
	bzero(data, sizeof(data));



	if (req_type == 0) /*snmpget*/
	{
		char line [128];
		FILE *version, *hostname, *uptime, *snmpd_conf;
		char sysdescr_string[40];

		NullNextOid(next_oid);
		OidCopier(next_oid, req_oid, oid_length);/*return the same oid in next_oid*/

		/*printf("[SystemFn] NextOId: %d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d.%d\n", 
		  next_oid[0], next_oid[1], next_oid[2], next_oid[3],
		  next_oid[4], next_oid[5], next_oid[6], next_oid[7],
		  next_oid[8], next_oid[9], next_oid[10], next_oid[11],
		  next_oid[12], next_oid[13], next_oid[14], next_oid[15]);  */
		switch (req_oid[7])
		{
			case 1: /*sysDescr*/
				SysDescrFn(data);
				break;	

			case 2: /*sysObjectId: since this is not a commercially registered software
					  we shall return 0 here */
				sprintf(data, "%d", 1);
				break;

			case 3: /*sysUpTime*/
				SysUptimeFn(data);
				break;

			case 4: /*sysContact*/
				SysContactFnRO(data);
				break;

			case 5: /*sysName*/
				SysNameFn(data);
				break;

			case 6: /*sysLocation*/
				SysLocationFnRO(data);
				break;

			case 7: /*sysServices*/
				sprintf(data, "%d", 7); /* this is an application box*/
				break;

			default:
				sprintf(data, "%d", -1); /* incorrect OID */
				break;
		}
	}  /*end of if (req_tyhpe==0) */

	else if (req_type == SNMPGETNEXT) /*snmpgetnext*/
	{
		char line [128];
		FILE *version, *hostname, *uptime, *snmpd_conf;
		char sysdescr_string[40];

		NullNextOid(next_oid); /*make allof 'em -1 's */

		/*if we did not get the whole OID length : */

		//printf("[SystemFn] oid_l %d", oid_length);
		if (oid_length < 36) /* u_long = 4B, oid = 9 x 4 = 36 */
		{
			/* we just return the data for first oid */

			SysDescrFn(data);
			OidCopier(next_oid, c_sysDescr0, sizeof(c_sysDescr0));
		} 

		else /* length is full length */
		{

			switch (req_oid[7])
			{
				case 1: /*sysDescr, return sysObjectId*/
					sprintf(data, "%d", 1);
					OidCopier(next_oid, c_sysObjectID0, sizeof(c_sysObjectID0));
					break;	

				case 2: /*sysObjectId: return sysUptime*/
					SysUptimeFn(data);
					OidCopier(next_oid, c_sysUpTime0, sizeof(c_sysUpTime0));
					break;

				case 3: /*sysUpTime return sysContact*/
					SysContactFnRO(data);
					OidCopier(next_oid, c_sysContact0, sizeof(c_sysContact0));
					break;

				case 4: /*sysContact return sysname*/
					SysNameFn(data);
					OidCopier(next_oid, c_sysName0, sizeof(c_sysName0));
					break;

				case 5: /*sysName return syslocation*/
					SysLocationFnRO(data);
					OidCopier(next_oid, c_sysLocation0, sizeof(c_sysLocation0));
					break;

				case 6: /*sysLocation return sysservices*/
					sprintf(data, "%d", 7); /* this is an application box*/
					OidCopier(next_oid, c_sysServices0, sizeof(c_sysServices0));
					break;

				case 7: /*sysServices*/
					sprintf(data, "%d", -2); /* end of subtree */
					break;

				default: /*incorrect OID */
					sprintf(data, "%d", -1); /* end of subtree */
					break;

			} /*end of switch */

		} /*end of else (length is full length) */

	} /*end of getnext*/
	else if (req_type == SNMPSET) /*snmpset*/
	{
		NullNextOid(next_oid);

		OidCopier(next_oid, req_oid, sizeof(next_oid));
		if (!((req_oid[7] == 4) || (req_oid[7] == 6)))
		{
			sprintf(data, "%d", -1); /* request made on bad oid*/
			return;
		}

		switch (req_oid[7])
		{
			case 4:
				SysContactFnRW(data, set_val);
				OidCopier(next_oid, c_sysContact0, sizeof(c_sysContact0));
				break;
			case 6:
				SysLocationFnRW(data,set_val);
				OidCopier(next_oid, c_sysLocation0, sizeof(c_sysLocation0));
				break;
		}
	}

} /*end of SystemFn() */


int SysContactFnRW(char *data, char *set_val) 
{
	FILE *snmpd_cfg;
	char line2[139];
	char line2_tmp[139];
	char *pch;
	char *blanco=" ";
	int seekpos = 0;
	int line_number = 0;
	int get_len, i, offset1, offset2;
	int c=0;

	//printf("string length set_val %d\n", strlen(set_val));
	set_val[strlen(set_val)] = '\0';
	if (!(snmpd_cfg = fopen("etc/syscontact.cfg", "r+")))
	{
		perror("[SysContactFnRW]: cannot open etc/syscontact.cfg");
		return;
	}

	while (line2 == fgets(line2, sizeof(line2), snmpd_cfg))
	{
		strcpy(line2_tmp, line2);
		pch = strtok (line2_tmp, ":");
		offset1 = ftell(snmpd_cfg);
		if (strcmp(pch,"sysContact") == 0)
		{
			pch = strtok(NULL, "\0");
			/*count the number of chars in pch*/
			for (i=0; (i < sizeof(line2)); i++)
			{
				//printf("pch_i is %c", pch[i]);
				if (pch[i] == '\0')
				{

					break;
				} else
					c++;
			}
			if	(fseek(snmpd_cfg, offset1 - c, SEEK_SET))
				perror("couldnt fseek");

			//printf("thus the start should be: %d\n", ftell(snmpd_cfg));

			if (!(fputs(set_val, snmpd_cfg) >0))
				perror ("couldn't write");

			/*we still have to fill the rest of it with null*/
			//printf("and now, we're at: %d\n", ftell(snmpd_cfg));

			while(ftell(snmpd_cfg)+1 < offset1)
				fputc(*blanco, snmpd_cfg);

			fputc('\n', snmpd_cfg);

			break; /*out of the while loop */
		}

	}

	if (pch[strlen(pch)-1] == '\n')
		pch[strlen(pch)-1] = '\0'; 



	strcpy(data, set_val);
	fclose(snmpd_cfg);
	1;

}

int SysLocationFnRW(char *data, char *set_val) 
{
	FILE *snmpd_cfg;
	char line2[139];
	char line2_tmp[139];
	char *pch;
	char *blanco=" ";
	int seekpos = 0;
	int line_number = 0;
	int get_len, i, offset1, offset2;
	int c=0;

	//printf("string length set_val %d\n", strlen(set_val));
	set_val[strlen(set_val)] = '\0';
	if (!(snmpd_cfg = fopen("etc/syslocation.cfg", "r+")))
	{
		perror("[SysLocationFnRW]: cannot open etc/syslocation.cfg");
		return;
	}

	while (line2 == fgets(line2, sizeof(line2), snmpd_cfg))
	{
		strcpy(line2_tmp, line2);
		pch = strtok (line2_tmp, ":");
		offset1 = ftell(snmpd_cfg);
		if (strcmp(pch,"sysLocation") == 0)
		{
			pch = strtok(NULL, "\0");
			/*count the number of chars in pch*/
			for (i=0; (i < sizeof(line2)); i++)
			{
				//printf("pch_i is %c", pch[i]);
				if (pch[i] == '\0')
				{

					break;
				} else
					c++;
			}
			if	(fseek(snmpd_cfg, offset1 - c, SEEK_SET))
				perror("couldnt fseek");

			//printf("thus the start should be: %d\n", ftell(snmpd_cfg));

			if (!(fputs(set_val, snmpd_cfg) >0))
				perror ("couldn't write");

			/*we still have to fill the rest of it with null*/
			//printf("and now, we're at: %d\n", ftell(snmpd_cfg));

			while(ftell(snmpd_cfg)+1 < offset1)
				fputc(*blanco, snmpd_cfg);

			fputc('\n', snmpd_cfg);

			break; /*out of the while loop */
		}

	}

	if (pch[strlen(pch)-1] == '\n')
		pch[strlen(pch)-1] = '\0'; 



	strcpy(data, set_val);
	fclose(snmpd_cfg);
	1;

}

